# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import FrozenSet, Union

import libcst as cst
from libcst.metadata import QualifiedName, QualifiedNameProvider, QualifiedNameSource

from fixit import CstLintRule, InvalidTestCase as Invalid, ValidTestCase as Valid


class CompareSingletonPrimitivesByIsRule(CstLintRule):
    """
    Enforces the use of `is` and `is not` in comparisons to singleton primitives (None, True, False) rather than == and !=.
    The == operator checks equality, when in this scenario, we want to check identity.
    See Flake8 rules E711 (https://www.flake8rules.com/rules/E711.html) and E712 (https://www.flake8rules.com/rules/E712.html).
    """

    MESSAGE: str = (
        "Comparisons to singleton primitives should not be done with == or !=, as they check equality rather than identiy."
        + " Use `is` or `is not` instead."
    )
    METADATA_DEPENDENCIES = (QualifiedNameProvider,)
    VALID = [
        Valid("if x: pass"),
        Valid("if not x: pass"),
        Valid("x is True"),
        Valid("x is False"),
        Valid("x is None"),
        Valid("x is not None"),
        Valid("x is True is not y"),
        Valid("y is None is not x"),
        Valid("None is y"),
        Valid("True is x"),
        Valid("False is x"),
        Valid("x == 2"),
        Valid("2 != x"),
    ]
    INVALID = [
        Invalid(
            code="x != True",
            expected_replacement="x is not True",
        ),
        Invalid(
            code="x != False",
            expected_replacement="x is not False",
        ),
        Invalid(
            code="x == False",
            expected_replacement="x is False",
        ),
        Invalid(
            code="x == None",
            expected_replacement="x is None",
        ),
        Invalid(
            code="x != None",
            expected_replacement="x is not None",
        ),
        Invalid(
            code="False == x",
            expected_replacement="False is x",
        ),
        Invalid(
            code="x is True == y",
            expected_replacement="x is True is y",
        ),
    ]

    QUALIFIED_SINGLETON_PRIMITIVES: FrozenSet[QualifiedName] = frozenset(
        {
            QualifiedName(name=f"builtins.{name}", source=QualifiedNameSource.BUILTIN)
            for name in ("True", "False", "None")
        }
    )

    def visit_Comparison(self, node: cst.Comparison) -> None:
        # Initialize the needs_report flag as False to begin with
        needs_report = False
        left_comp = node.left
        altered_comparisons = []
        for target in node.comparisons:
            operator, right_comp = target.operator, target.comparator
            if isinstance(operator, (cst.Equal, cst.NotEqual)) and (
                not self.QUALIFIED_SINGLETON_PRIMITIVES.isdisjoint(
                    self.get_metadata(QualifiedNameProvider, left_comp, set())
                )
                or not self.QUALIFIED_SINGLETON_PRIMITIVES.isdisjoint(
                    self.get_metadata(QualifiedNameProvider, right_comp, set())
                )
            ):
                needs_report = True
                altered_comparisons.append(
                    target.with_changes(operator=self.alter_operator(operator))
                )
            else:
                altered_comparisons.append(target)
            # Continue the check down the line of comparisons, if more than one
            left_comp = right_comp

        if needs_report:
            self.report(
                node, replacement=node.with_changes(comparisons=altered_comparisons)
            )

    def alter_operator(
        self, original_op: Union[cst.Equal, cst.NotEqual]
    ) -> Union[cst.Is, cst.IsNot]:
        return (
            cst.IsNot(
                whitespace_before=original_op.whitespace_before,
                whitespace_after=original_op.whitespace_after,
            )
            if isinstance(original_op, cst.NotEqual)
            else cst.Is(
                whitespace_before=original_op.whitespace_before,
                whitespace_after=original_op.whitespace_after,
            )
        )



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import tokenize
from io import BytesIO

from libcst.testing.utils import UnitTest

from fixit.common.comments import CommentInfo
from fixit.common.utils import dedent_with_lstrip


class CommentInfoTest(UnitTest):
    def test_comment_info(self) -> None:
        # A comment on a line with no other leading tokens is a "comment on own line".
        # In contrast, trailing comments come after other tokens on the same line.
        code = dedent_with_lstrip(
            """
            # comment on own line
            # this is a
            # multiline comment
            def fn():
                # comment on own line
                fn2()
                fn3()  # trailing comment
                fn4()
                # comment on own line
            """
        )
        tokens = tokenize.tokenize(BytesIO(code.encode("utf-8")).readline)
        result = CommentInfo.compute(tokens=tokens)
        # The set of all comments includes both comments on their own line and trailing
        # comments.
        self.assertEqual([tok.start[0] for tok in result.comments], [1, 2, 3, 5, 7, 9])
        # `comments_on_own_line` is a subset of all comments
        self.assertEqual(
            [tok.start[0] for tok in result.comments_on_own_line], [1, 2, 3, 5, 9]
        )



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
All of the ignore logic for the lint engine.
"""

import tokenize
from collections import defaultdict
from dataclasses import dataclass
from enum import Enum
from typing import (
    Collection,
    Dict,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)

from fixit.common.base import CstLintRule
from fixit.common.comments import CommentInfo
from fixit.common.config import (
    FLAKE8_NOQA_FILE,
    LINT_IGNORE_REGEXP,
    NOQA_FILE_RULE,
    NOQA_INLINE_REGEXP,
)
from fixit.common.insert_suppressions import BODY_PREFIX_WITH_SPACE
from fixit.common.line_mapping import LineMappingInfo
from fixit.common.pseudo_rule import PseudoLintRule
from fixit.common.report import BaseLintRuleReport


_LintRuleT = Union[Type[CstLintRule], Type[PseudoLintRule]]


class AllRulesType(Enum):
    ALL_RULES = 0


IgnoredRules = Union[Collection[str], AllRulesType]


class SuppressionCommentKind(Enum):
    NOQA = "noqa"
    LINT_IGNORE = "lint-ignore"  # also includes lint-fixme comments


class SuppressionComment:
    ignored_rules: IgnoredRules
    # a lint-fixme or lint-ignore comment can span multiple lines, so it may be composed
    # of multiple tokens.
    tokens: Sequence[tokenize.TokenInfo]
    used_by: List[BaseLintRuleReport]
    kind: str
    reason: Optional[str]

    def __init__(
        self,
        ignored_rules: IgnoredRules,
        tokens: Sequence[tokenize.TokenInfo],
        kind: str,
        reason: Optional[str] = None,
    ) -> None:
        self.ignored_rules = ignored_rules
        self.tokens = tokens
        self.used_by = []
        self.kind = kind
        self.reason = reason

    def should_ignore_report(self, report: BaseLintRuleReport) -> bool:
        ignored_rules = self.ignored_rules
        return isinstance(ignored_rules, AllRulesType) or report.code in ignored_rules

    def mark_used_by(self, report: BaseLintRuleReport) -> None:
        self.used_by.append(report)

    def __repr__(self) -> str:
        return "\n".join(t.string for t in self.tokens)


# Loosely based on flake8's parse_comma_separated_list
# https://gitlab.com/pycqa/flake8/blob/52d88d8ca7208d1edc554b41a/src/flake8/utils.py#L17
# Used for parsing `# lint-fixme`, `# lint-ignore`, `# noqa` and `# noqa-file`
def _parse_comma_separated_rules(rules_str: Optional[str]) -> IgnoredRules:
    if rules_str is None:
        return AllRulesType.ALL_RULES
    item_gen = (c.strip() for c in rules_str.split(","))
    rules_list = [item for item in item_gen if item]
    return rules_list if rules_list else AllRulesType.ALL_RULES


@dataclass(frozen=True)
class GlobalIgnoreInfo:
    # Rules that are ignored on every line of the file (e.g. due to a `# noqa-file`).
    globally_ignored_rules: IgnoredRules

    def should_evaluate_rule(self, rule: _LintRuleT) -> bool:
        # TODO: We need a way to map lint codes back to the rules they come from so that
        # we can avoid executing them.
        return not isinstance(self.globally_ignored_rules, AllRulesType)

    def should_ignore_report(self, report: BaseLintRuleReport) -> bool:
        globally_ignored_rules = self.globally_ignored_rules
        return (
            isinstance(globally_ignored_rules, AllRulesType)
            or report.code in globally_ignored_rules
        )

    @staticmethod
    def compute(*, comment_info: CommentInfo, use_noqa: bool) -> "GlobalIgnoreInfo":
        if not use_noqa:
            return GlobalIgnoreInfo(globally_ignored_rules=set())
        ignored_rules = set()
        for tok in comment_info.comments_on_own_line:
            if FLAKE8_NOQA_FILE.fullmatch(tok.string):
                # For backwards compatibility
                return GlobalIgnoreInfo(globally_ignored_rules=AllRulesType.ALL_RULES)
            file_rule_match = NOQA_FILE_RULE.fullmatch(tok.string)
            if file_rule_match:
                newly_ignored_rules = _parse_comma_separated_rules(
                    file_rule_match.group("codes")
                )
                if isinstance(newly_ignored_rules, AllRulesType):
                    raise ValueError("A `# noqa-file` must specify codes to ignore")
                ignored_rules.update(newly_ignored_rules)
        return GlobalIgnoreInfo(globally_ignored_rules=ignored_rules)


@dataclass(frozen=True)
class LocalIgnoreInfo:
    local_suppression_comments: Collection[SuppressionComment]
    # Maps logical line numbers to the relevant suppression comments for that line.
    local_suppression_comments_by_line: Mapping[int, Sequence[SuppressionComment]]
    # We use this to find the next non-empty logical line before looking up
    # locally_ignored_rules.
    line_mapping_info: LineMappingInfo

    def should_ignore_report(self, report: BaseLintRuleReport) -> bool:
        logical_line = self.line_mapping_info.physical_to_logical[report.line]
        suppression_comments = self.local_suppression_comments_by_line.get(
            logical_line, []
        )
        for comment in suppression_comments:
            if comment.should_ignore_report(report):
                comment.mark_used_by(report)
                return True
        return False

    @staticmethod
    def get_all_tokens_and_full_reason(
        tokens: List[tokenize.TokenInfo],
        comment_lines: Iterator[tokenize.TokenInfo],
        end_line: int,
        reason: Optional[str],
    ) -> Tuple[List[tokenize.TokenInfo], Optional[str], Optional[tokenize.TokenInfo]]:
        next_comment_line = next(comment_lines, None)
        while next_comment_line is not None and next_comment_line.start[0] < end_line:
            if not next_comment_line.string.startswith(BODY_PREFIX_WITH_SPACE):
                break
            reason_contd = next_comment_line.string.split(BODY_PREFIX_WITH_SPACE, 1)[1]
            reason = (
                " ".join([reason, reason_contd]) if reason is not None else reason_contd
            )
            tokens.append(next_comment_line)
            next_comment_line = next(comment_lines, None)

        return tokens, reason, next_comment_line

    @staticmethod
    def compute(
        *, comment_info: CommentInfo, line_mapping_info: LineMappingInfo, use_noqa: bool
    ) -> "LocalIgnoreInfo":
        local_suppression_comments: List[SuppressionComment] = []
        local_suppression_comments_by_line: Dict[
            int, List[SuppressionComment]
        ] = defaultdict(list)

        # New `# lint-fixme` and `# lint-ignore` comments. These are preferred over
        # legacy `# noqa` comments.
        comments_on_own_line_iter = iter(comment_info.comments_on_own_line)
        next_comment_line: Optional[tokenize.TokenInfo] = next(
            comments_on_own_line_iter, None
        )

        while next_comment_line is not None:
            match = LINT_IGNORE_REGEXP.fullmatch(next_comment_line.string)
            if match is not None:
                # We are at the *start* of a suppression comment. There may be more physical lines
                # to the comment. We assume any lines starting with `# lint: ` are a continuation.
                start_line = next_comment_line.start[0]
                end_line = line_mapping_info.get_next_non_empty_logical_line(start_line)
                assert end_line is not None, "Failed to get next non-empty logical line"

                (
                    tokens,
                    reason,
                    next_comment_line,
                ) = LocalIgnoreInfo.get_all_tokens_and_full_reason(
                    [next_comment_line],
                    comments_on_own_line_iter,
                    end_line,
                    match.group("reason"),
                )

                codes = _parse_comma_separated_rules(match.group("codes"))
                # Construct the SuppressionComment with all the information.
                comment = SuppressionComment(codes, tokens, match.group(1), reason)

                local_suppression_comments.append(comment)

                for tok in tokens:
                    local_suppression_comments_by_line[tok.start[0]].append(comment)

                # Finally we want to map the suppressed line of code to this suppression comment.
                local_suppression_comments_by_line[end_line].append(comment)
            else:
                next_comment_line = next(comments_on_own_line_iter, None)

        # Legacy inline `# noqa` comments. This matches flake8's behavior.
        # Process these after `# lint-ignore` comments, because in the case of duplicate
        # or overlapping ignores, we'd prefer to mark the noqa as unused, instead of the
        # more modern `# lint-ignore` comment.
        if use_noqa is True:
            for tok in comment_info.comments:
                match = NOQA_INLINE_REGEXP.search(tok.string)
                if match:
                    normalized_line = line_mapping_info.physical_to_logical[
                        tok.start[0]
                    ]
                    codes = _parse_comma_separated_rules(match.group("codes"))
                    comment = SuppressionComment(codes, [tok], kind="noqa")
                    local_suppression_comments.append(comment)
                    local_suppression_comments_by_line[normalized_line].append(comment)

        return LocalIgnoreInfo(
            local_suppression_comments,
            dict(local_suppression_comments_by_line),  # no longer a defaultdict
            line_mapping_info,
        )


@dataclass(frozen=True)
class IgnoreInfo:
    global_ignore_info: GlobalIgnoreInfo
    local_ignore_info: LocalIgnoreInfo
    suppression_comments: Collection[SuppressionComment]

    def should_evaluate_rule(self, rule: _LintRuleT) -> bool:
        """
        Call this before evaluating lint rules to filter out rules that will be entirely
        ignored.
        """
        return self.global_ignore_info.should_evaluate_rule(rule)

    def should_ignore_report(self, report: BaseLintRuleReport) -> bool:
        return self.global_ignore_info.should_ignore_report(
            report
        ) or self.local_ignore_info.should_ignore_report(report)

    @staticmethod
    def compute(
        *, comment_info: CommentInfo, line_mapping_info: LineMappingInfo, use_noqa: bool
    ) -> "IgnoreInfo":
        global_ignore_info = GlobalIgnoreInfo.compute(
            comment_info=comment_info, use_noqa=use_noqa
        )
        local_ignore_info = LocalIgnoreInfo.compute(
            comment_info=comment_info,
            line_mapping_info=line_mapping_info,
            use_noqa=use_noqa,
        )
        return IgnoreInfo(
            global_ignore_info,
            local_ignore_info,
            # TODO: compute global suppression comments and merge them here
            local_ignore_info.local_suppression_comments,
        )



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import libcst as cst

from fixit import CstLintRule, InvalidTestCase as Invalid, ValidTestCase as Valid


class ComparePrimitivesByEqualRule(CstLintRule):
    """
    Enforces the use of ``==`` and ``!=`` in comparisons to primitives rather than ``is`` and ``is not``.
    The ``==`` operator checks equality (https://docs.python.org/3/reference/datamodel.html#object.__eq__),
    while ``is`` checks identity (https://docs.python.org/3/reference/expressions.html#is).
    """

    MESSAGE = "Don't use `is` or `is not` to compare primitives, as they compare references. Use == or != instead."
    VALID = [
        Valid("a == 1"),
        Valid("a == '1'"),
        Valid("a != '1'"),
        Valid("'3' == '1'"),
        Valid("3 == '1'"),
        Valid("3 > 2 > 1"),
        Valid("3 > 2 > '1'"),
        Valid("a is b > 1"),
        Valid("a is b is c"),
        Valid("1 > b is c"),
    ]
    INVALID = [
        Invalid("a is 1", expected_replacement="a == 1"),
        Invalid("a is '1'", expected_replacement="a == '1'"),
        Invalid(
            "a is f'1{b}'",
            expected_replacement="a == f'1{b}'",
        ),
        Invalid(
            "a is not f'1{d}'",
            expected_replacement="a != f'1{d}'",
        ),
        Invalid("1 is a", expected_replacement="1 == a"),
        Invalid(
            "'2' > '1' is a",
            expected_replacement="'2' > '1' == a",
        ),
        Invalid(
            "3 > a is 2",
            expected_replacement="3 > a == 2",
        ),
        Invalid(
            "1  is   2",
            expected_replacement="1  ==   2",
        ),
    ]
    PRIMITIVES = (cst.BaseNumber, cst.BaseString)

    def visit_Comparison(self, node: cst.Comparison) -> None:
        prev_comparator = node.left
        for target in node.comparisons:
            op, comparator = target.operator, target.comparator
            if isinstance(op, (cst.Is, cst.IsNot)):
                if (
                    prev_comparator and isinstance(prev_comparator, self.PRIMITIVES)
                ) or isinstance(comparator, self.PRIMITIVES):
                    self.report(node, replacement=self.replace_operators(node))
                    return
            prev_comparator = comparator

    def replace_operators(self, node: cst.Comparison) -> cst.Comparison:
        prev_comparator = node.left
        comparisons = []
        for target in node.comparisons:
            op, comparator = target.operator, target.comparator
            if isinstance(op, (cst.Is, cst.IsNot)):
                if isinstance(prev_comparator, self.PRIMITIVES) or isinstance(
                    comparator, self.PRIMITIVES
                ):
                    target = target.with_changes(
                        operator=cst.Equal(
                            whitespace_before=op.whitespace_before,
                            whitespace_after=op.whitespace_after,
                        )
                        if isinstance(op, cst.Is)
                        else cst.NotEqual(
                            whitespace_before=op.whitespace_before,
                            whitespace_after=op.whitespace_after,
                        )
                    )
            comparisons.append(target)
            prev_comparator = comparator

        return node.with_changes(comparisons=comparisons)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import re
from typing import cast

import libcst as cst
import libcst.matchers as m

from fixit import (
    CstContext,
    CstLintRule,
    InvalidTestCase as Invalid,
    ValidTestCase as Valid,
)


class UseClassNameAsCodeRule(CstLintRule):
    """
    Meta lint rule which checks that codes of lint rules are migrated to new format in lint rule class definitions.
    """

    MESSAGE = "`IG`-series codes are deprecated. Use class name as code instead."
    VALID = [
        Valid(
            """
        MESSAGE = "This is a message"
        """
        ),
        Valid(
            """
        from fixit.common.base import CstLintRule
        class FakeRule(CstLintRule):
            MESSAGE = "This is a message"
        """
        ),
        Valid(
            """
        from fixit.common.base import CstLintRule
        class FakeRule(CstLintRule):
            INVALID = [
                Invalid(
                    code=""
                )
            ]
        """
        ),
    ]
    INVALID = [
        Invalid(
            code="""
            MESSAGE = "IG90000 Message"
            """,
            expected_replacement="""
            MESSAGE = "Message"
            """,
        ),
        Invalid(
            code="""
            from fixit.common.base import CstLintRule
            class FakeRule(CstLintRule):
                INVALID = [
                    Invalid(
                        code="",
                        kind="IG000"
                    )
                ]
            """,
            expected_replacement="""
            from fixit.common.base import CstLintRule
            class FakeRule(CstLintRule):
                INVALID = [
                    Invalid(
                        code="",
                        )
                ]
            """,
        ),
    ]

    def __init__(self, context: CstContext) -> None:
        super().__init__(context)
        self.inside_invalid_call: bool = False

    def visit_SimpleString(self, node: cst.SimpleString) -> None:
        matched = re.match(r"^(\'|\")(?P<igcode>IG\d+ )\S", node.value)

        if matched is not None:
            replacement_string = node.value.replace(matched.group("igcode"), "", 1)
            self.report(
                node,
                self.MESSAGE,
                replacement=node.with_changes(value=replacement_string),
            )

    def visit_Call(self, node: cst.Call) -> None:
        func = node.func
        if m.matches(func, m.Name()):
            func = cast(cst.Name, func)
            if func.value == "Invalid":
                self.inside_invalid_call = True

    def leave_Call(self, original_node: cst.Call) -> None:
        func = original_node.func
        if m.matches(func, m.Name()):
            func = cast(cst.Name, func)
            if func.value == "Invalid":
                self.inside_invalid_call = False

    def visit_Arg(self, node: cst.Arg) -> None:
        # Remove `kind` arguments to Invalid test cases as they are no longer needed
        if self.inside_invalid_call:
            arg_value = node.value
            if m.matches(arg_value, m.SimpleString()):
                arg_value = cast(cst.SimpleString, arg_value)
                string_value = arg_value.value
                matched = re.match(r"^(\'|\")(?P<igcode>IG\d+)(\'|\")\Z", string_value)
                if matched:
                    self.report(
                        node,
                        self.MESSAGE,
                        replacement=cst.RemovalSentinel.REMOVE,
                    )



# Copyright (c) Meta Platforms, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
import sys
from pathlib import Path
from typing import Iterable, Optional

import click

from fixit import __version__

from .api import fixit_paths
from .types import Options


@click.group()
@click.pass_context
@click.version_option(__version__, "--version", "-V", prog_name="fixit")
@click.option(
    "--debug/--quiet", is_flag=True, default=None, help="Increase decrease verbosity"
)
@click.option(
    "--config-file",
    "-c",
    type=click.Path(dir_okay=False, exists=True, path_type=Path),
    default=None,
    help="Override default config file search behavior",
)
def main(
    ctx: click.Context,
    debug: Optional[bool],
    config_file: Optional[Path],
):
    level = logging.WARNING
    if debug is not None:
        level = logging.DEBUG if debug else logging.ERROR
    logging.basicConfig(level=level, stream=sys.stderr)

    ctx.obj = Options(
        debug=debug,
        config_file=config_file,
    )


@main.command()
@click.pass_context
@click.argument("paths", nargs=-1, type=click.Path(path_type=Path))
def lint(
    ctx: click.Context,
    paths: Iterable[Path],
):
    """
    lint one or more paths and return suggestions
    """
    for result in fixit_paths(paths):
        print(result)


@main.command()
@click.pass_context
@click.argument("paths", nargs=-1, type=click.Path(path_type=Path))
def fix(
    ctx: click.Context,
    paths: Iterable[Path],
):
    """
    lint and autofix one or more files and return results
    """
    ctx.fail("not implemented yet")



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import libcst as cst
import libcst.matchers as m
from libcst.helpers import ensure_type

from fixit import CstLintRule, InvalidTestCase as Invalid, ValidTestCase as Valid


class UseAssertInRule(CstLintRule):
    """
    Discourages use of ``assertTrue(x in y)`` and ``assertFalse(x in y)``
    as it is deprecated (https://docs.python.org/3.8/library/unittest.html#deprecated-aliases).
    Use ``assertIn(x, y)`` and ``assertNotIn(x, y)``) instead.
    """

    MESSAGE: str = (
        "Use assertIn/assertNotIn instead of assertTrue/assertFalse for inclusion check.\n"
        + "See https://docs.python.org/3/library/unittest.html#unittest.TestCase.assertIn)"
    )

    VALID = [
        Valid("self.assertIn(a, b)"),
        Valid("self.assertIn(f(), b)"),
        Valid("self.assertIn(f(x), b)"),
        Valid("self.assertIn(f(g(x)), b)"),
        Valid("self.assertNotIn(a, b)"),
        Valid("self.assertNotIn(f(), b)"),
        Valid("self.assertNotIn(f(x), b)"),
        Valid("self.assertNotIn(f(g(x)), b)"),
    ]

    INVALID = [
        Invalid(
            "self.assertTrue(a in b)",
            expected_replacement="self.assertIn(a, b)",
        ),
        Invalid(
            "self.assertTrue(f() in b)",
            expected_replacement="self.assertIn(f(), b)",
        ),
        Invalid(
            "self.assertTrue(f(x) in b)",
            expected_replacement="self.assertIn(f(x), b)",
        ),
        Invalid(
            "self.assertTrue(f(g(x)) in b)",
            expected_replacement="self.assertIn(f(g(x)), b)",
        ),
        Invalid(
            "self.assertTrue(a not in b)",
            expected_replacement="self.assertNotIn(a, b)",
        ),
        Invalid(
            "self.assertTrue(not a in b)",
            expected_replacement="self.assertNotIn(a, b)",
        ),
        Invalid(
            "self.assertFalse(a in b)",
            expected_replacement="self.assertNotIn(a, b)",
        ),
    ]

    def visit_Call(self, node: cst.Call) -> None:
        # Todo: Make use of single extract instead of having several
        # if else statemenets to make the code more robust and readable.
        if m.matches(
            node,
            m.Call(
                func=m.Attribute(value=m.Name("self"), attr=m.Name("assertTrue")),
                args=[
                    m.Arg(
                        m.Comparison(comparisons=[m.ComparisonTarget(operator=m.In())])
                    )
                ],
            ),
        ):
            # self.assertTrue(a in b) -> self.assertIn(a, b)
            new_call = node.with_changes(
                func=cst.Attribute(value=cst.Name("self"), attr=cst.Name("assertIn")),
                args=[
                    cst.Arg(ensure_type(node.args[0].value, cst.Comparison).left),
                    cst.Arg(
                        ensure_type(node.args[0].value, cst.Comparison)
                        .comparisons[0]
                        .comparator
                    ),
                ],
            )
            self.report(node, replacement=new_call)
        else:
            # ... -> self.assertNotIn(a, b)
            matched, arg1, arg2 = False, None, None
            if m.matches(
                node,
                m.Call(
                    func=m.Attribute(value=m.Name("self"), attr=m.Name("assertTrue")),
                    args=[
                        m.Arg(
                            m.UnaryOperation(
                                operator=m.Not(),
                                expression=m.Comparison(
                                    comparisons=[m.ComparisonTarget(operator=m.In())]
                                ),
                            )
                        )
                    ],
                ),
            ):
                # self.assertTrue(not a in b) -> self.assertNotIn(a, b)
                matched = True
                arg1 = cst.Arg(
                    ensure_type(
                        ensure_type(node.args[0].value, cst.UnaryOperation).expression,
                        cst.Comparison,
                    ).left
                )
                arg2 = cst.Arg(
                    ensure_type(
                        ensure_type(node.args[0].value, cst.UnaryOperation).expression,
                        cst.Comparison,
                    )
                    .comparisons[0]
                    .comparator
                )
            elif m.matches(
                node,
                m.Call(
                    func=m.Attribute(value=m.Name("self"), attr=m.Name("assertTrue")),
                    args=[
                        m.Arg(m.Comparison(comparisons=[m.ComparisonTarget(m.NotIn())]))
                    ],
                ),
            ):
                # self.assertTrue(a not in b) -> self.assertNotIn(a, b)
                matched = True
                arg1 = cst.Arg(ensure_type(node.args[0].value, cst.Comparison).left)
                arg2 = cst.Arg(
                    ensure_type(node.args[0].value, cst.Comparison)
                    .comparisons[0]
                    .comparator
                )
            elif m.matches(
                node,
                m.Call(
                    func=m.Attribute(value=m.Name("self"), attr=m.Name("assertFalse")),
                    args=[
                        m.Arg(m.Comparison(comparisons=[m.ComparisonTarget(m.In())]))
                    ],
                ),
            ):
                # self.assertFalse(a in b) -> self.assertNotIn(a, b)
                matched = True
                arg1 = cst.Arg(ensure_type(node.args[0].value, cst.Comparison).left)
                arg2 = cst.Arg(
                    ensure_type(node.args[0].value, cst.Comparison)
                    .comparisons[0]
                    .comparator
                )

            if matched:
                new_call = node.with_changes(
                    func=cst.Attribute(
                        value=cst.Name("self"), attr=cst.Name("assertNotIn")
                    ),
                    args=[arg1, arg2],
                )
                self.report(node, replacement=new_call)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import cast, Dict, Optional, Sequence, Union

import libcst as cst
import libcst.matchers as m
from libcst.helpers import get_full_name_for_node
from libcst.metadata import TypeInferenceProvider

from fixit import CstLintRule, InvalidTestCase as Invalid, ValidTestCase as Valid


class AwaitAsyncCallRule(CstLintRule):
    """
    Enforces calls to coroutines are preceeded by the ``await`` keyword. Awaiting on a coroutine will execute it while
    simply calling a coroutine returns a coroutine object (https://docs.python.org/3/library/asyncio-task.html#coroutines).
    """

    MESSAGE: str = (
        "Async function call will only be executed with `await` statement. Did you forget to add `await`? "
        + "If you intend to not await, please add comment to disable this warning: # lint-fixme: AwaitAsyncCallRule "
    )

    METADATA_DEPENDENCIES = (TypeInferenceProvider,)

    VALID = [
        Valid(
            """
            async def async_func():
                await async_foo()
            """
        ),
        Valid(
            """
            def foo(): pass
            foo()
            """
        ),
        Valid(
            """
            async def foo(): pass
            async def bar():
                await foo()
            """
        ),
        Valid(
            """
            async def foo(): pass
            async def bar():
                x = await foo()
            """
        ),
        Valid(
            """
            async def foo() -> bool: pass
            async def bar():
                while not await foo(): pass
            """
        ),
        Valid(
            """
            import asyncio
            async def foo(): pass
            asyncio.run(foo())
            """
        ),
    ]
    INVALID = [
        Invalid(
            """
            async def foo(): pass
            async def bar():
                foo()
            """,
            expected_replacement="""
            async def foo(): pass
            async def bar():
                await foo()
            """,
        ),
        Invalid(
            """
            class Foo:
                async def _attr(self): pass
            obj = Foo()
            obj._attr
            """,
            expected_replacement="""
            class Foo:
                async def _attr(self): pass
            obj = Foo()
            await obj._attr
            """,
        ),
        Invalid(
            """
            class Foo:
                async def _method(self): pass
            obj = Foo()
            obj._method()
            """,
            expected_replacement="""
            class Foo:
                async def _method(self): pass
            obj = Foo()
            await obj._method()
            """,
        ),
        Invalid(
            """
            class Foo:
                async def _method(self): pass
            obj = Foo()
            result = obj._method()
            """,
            expected_replacement="""
            class Foo:
                async def _method(self): pass
            obj = Foo()
            result = await obj._method()
            """,
        ),
        Invalid(
            """
            class Foo:
                async def bar(): pass
            class NodeUser:
                async def get():
                    do_stuff()
                    return Foo()
            user = NodeUser.get().bar()
            """,
            expected_replacement="""
            class Foo:
                async def bar(): pass
            class NodeUser:
                async def get():
                    do_stuff()
                    return Foo()
            user = await NodeUser.get().bar()
            """,
        ),
        Invalid(
            """
            class Foo:
                async def _attr(self): pass
            obj = Foo()
            attribute = obj._attr
            """,
            expected_replacement="""
            class Foo:
                async def _attr(self): pass
            obj = Foo()
            attribute = await obj._attr
            """,
        ),
        Invalid(
            code="""
            async def foo() -> bool: pass
            x = True
            if x and foo(): pass
            """,
            expected_replacement="""
            async def foo() -> bool: pass
            x = True
            if x and await foo(): pass
            """,
        ),
        Invalid(
            code="""
            async def foo() -> bool: pass
            x = True
            are_both_true = x and foo()
            """,
            expected_replacement="""
            async def foo() -> bool: pass
            x = True
            are_both_true = x and await foo()
            """,
        ),
        Invalid(
            """
            async def foo() -> bool: pass
            if foo():
                do_stuff()
            """,
            expected_replacement="""
            async def foo() -> bool: pass
            if await foo():
                do_stuff()
            """,
        ),
        Invalid(
            """
            async def foo() -> bool: pass
            if not foo():
                do_stuff()
            """,
            expected_replacement="""
            async def foo() -> bool: pass
            if not await foo():
                do_stuff()
            """,
        ),
        Invalid(
            """
            class Foo:
                async def _attr(self): pass
                def bar(self):
                    if self._attr: pass
            """,
            expected_replacement="""
            class Foo:
                async def _attr(self): pass
                def bar(self):
                    if await self._attr: pass
            """,
        ),
        Invalid(
            """
            class Foo:
                async def _attr(self): pass
                def bar(self):
                    if not self._attr: pass
            """,
            expected_replacement="""
            class Foo:
                async def _attr(self): pass
                def bar(self):
                    if not await self._attr: pass
            """,
        ),
        # Case where only cst.Attribute node's `attr` returns awaitable
        Invalid(
            """
            class Foo:
                async def _attr(self): pass
            def bar() -> Foo:
                return Foo()
            attribute = bar()._attr
            """,
            expected_replacement="""
            class Foo:
                async def _attr(self): pass
            def bar() -> Foo:
                return Foo()
            attribute = await bar()._attr
            """,
        ),
        # Case where only cst.Attribute node's `value` returns awaitable
        Invalid(
            """
            class Foo:
                def _attr(self): pass
            async def bar():
                await do_stuff()
                return Foo()
            attribute = bar()._attr
            """,
            expected_replacement="""
            class Foo:
                def _attr(self): pass
            async def bar():
                await do_stuff()
                return Foo()
            attribute = await bar()._attr
            """,
        ),
        Invalid(
            """
            async def bar() -> bool: pass
            while bar(): pass
            """,
            expected_replacement="""
            async def bar() -> bool: pass
            while await bar(): pass
            """,
        ),
        Invalid(
            """
            async def bar() -> bool: pass
            while not bar(): pass
            """,
            expected_replacement="""
            async def bar() -> bool: pass
            while not await bar(): pass
            """,
        ),
        Invalid(
            """
            class Foo:
                @classmethod
                async def _method(cls): pass
            Foo._method()
            """,
            expected_replacement="""
            class Foo:
                @classmethod
                async def _method(cls): pass
            await Foo._method()
            """,
        ),
        Invalid(
            """
            class Foo:
                @staticmethod
                async def _method(self): pass
            Foo._method()
            """,
            expected_replacement="""
            class Foo:
                @staticmethod
                async def _method(self): pass
            await Foo._method()
            """,
        ),
    ]

    @staticmethod
    def _is_awaitable_callable(annotation: str) -> bool:
        if not (
            annotation.startswith("typing.Callable")
            or annotation.startswith("typing.ClassMethod")
            or annotation.startswith("StaticMethod")
        ):
            # Exit early if this is not even a `typing.Callable` annotation.
            return False
        try:
            # Wrap this in a try-except since the type annotation may not be parse-able as a module.
            # If it is not parse-able, we know it's not what we are looking for anyway, so return `False`.
            parsed_ann = cst.parse_module(annotation)
        except Exception:
            return False
        # If passed annotation does not match the expected annotation structure for a `typing.Callable` with
        # typing.Coroutine as the return type, matched_callable_ann will simply be `None`.
        # The expected structure of an awaitable callable annotation from Pyre is: typing.Callable()[[...], typing.Coroutine[...]]
        matched_callable_ann: Optional[
            Dict[str, Union[Sequence[cst.CSTNode], cst.CSTNode]]
        ] = m.extract(
            parsed_ann,
            m.Module(
                body=[
                    m.SimpleStatementLine(
                        body=[
                            m.Expr(
                                value=m.Subscript(
                                    slice=[
                                        m.SubscriptElement(),
                                        m.SubscriptElement(
                                            slice=m.Index(
                                                value=m.Subscript(
                                                    value=m.SaveMatchedNode(
                                                        m.Attribute(),
                                                        "base_return_type",
                                                    )
                                                )
                                            )
                                        ),
                                    ],
                                )
                            )
                        ]
                    ),
                ]
            ),
        )
        if (
            matched_callable_ann is not None
            and "base_return_type" in matched_callable_ann
        ):
            base_return_type = get_full_name_for_node(
                cst.ensure_type(matched_callable_ann["base_return_type"], cst.CSTNode)
            )
            return (
                base_return_type is not None and base_return_type == "typing.Coroutine"
            )
        return False

    def _get_awaitable_replacement(self, node: cst.CSTNode) -> Optional[cst.CSTNode]:
        annotation = self.get_metadata(TypeInferenceProvider, node, None)
        if annotation is not None and (
            annotation.startswith("typing.Coroutine")
            or self._is_awaitable_callable(annotation)
        ):
            if isinstance(node, cst.BaseExpression):
                return cst.Await(expression=node)
        return None

    def _get_async_attr_replacement(self, node: cst.Attribute) -> Optional[cst.CSTNode]:
        value = node.value
        if m.matches(value, m.Call()):
            value = cast(cst.Call, value)
            value_replacement = self._get_async_call_replacement(value)
            if value_replacement is not None:
                return node.with_changes(value=value_replacement)
        return self._get_awaitable_replacement(node)

    def _get_async_call_replacement(self, node: cst.Call) -> Optional[cst.CSTNode]:
        func = node.func
        if m.matches(func, m.Attribute()):
            func = cast(cst.Attribute, func)
            attr_func_replacement = self._get_async_attr_replacement(func)
            if attr_func_replacement is not None:
                return node.with_changes(func=attr_func_replacement)
        return self._get_awaitable_replacement(node)

    def _get_async_expr_replacement(self, node: cst.CSTNode) -> Optional[cst.CSTNode]:
        if m.matches(node, m.Call()):
            node = cast(cst.Call, node)
            return self._get_async_call_replacement(node)
        elif m.matches(node, m.Attribute()):
            node = cast(cst.Attribute, node)
            return self._get_async_attr_replacement(node)
        elif m.matches(node, m.UnaryOperation(operator=m.Not())):
            node = cast(cst.UnaryOperation, node)
            replacement_expression = self._get_async_expr_replacement(node.expression)
            if replacement_expression is not None:
                return node.with_changes(expression=replacement_expression)
        elif m.matches(node, m.BooleanOperation()):
            node = cast(cst.BooleanOperation, node)
            maybe_left = self._get_async_expr_replacement(node.left)
            maybe_right = self._get_async_expr_replacement(node.right)
            if maybe_left is not None or maybe_right is not None:
                left_replacement = maybe_left if maybe_left is not None else node.left
                right_replacement = (
                    maybe_right if maybe_right is not None else node.right
                )
                return node.with_changes(left=left_replacement, right=right_replacement)
        return None

    def _maybe_autofix_node(self, node: cst.CSTNode, attribute_name: str) -> None:
        replacement_value = self._get_async_expr_replacement(
            getattr(node, attribute_name)
        )
        if replacement_value is not None:
            replacement = node.with_changes(**{attribute_name: replacement_value})
            self.report(node, replacement=replacement)

    def visit_If(self, node: cst.If) -> None:
        self._maybe_autofix_node(node, "test")

    def visit_While(self, node: cst.While) -> None:
        self._maybe_autofix_node(node, "test")

    def visit_Assign(self, node: cst.Assign) -> None:
        self._maybe_autofix_node(node, "value")

    def visit_Expr(self, node: cst.Expr) -> None:
        self._maybe_autofix_node(node, "value")



# Copyright (c) Meta Platforms, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest

if __name__ == "__main__":
    unittest.main("fixit.tests", verbosity=2)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import tokenize
from bisect import bisect_left
from dataclasses import dataclass
from typing import Container, Iterable, Mapping, Optional, Sequence


_LOGICAL_LINE_END_MARKERS: Container[int] = (
    tokenize.NL,
    tokenize.NEWLINE,
    tokenize.ENDMARKER,
)
_EMPTY_TOKENS: Container[int] = (
    tokenize.ENCODING,
    tokenize.COMMENT,
    tokenize.INDENT,
    tokenize.DEDENT,
)


@dataclass(frozen=True)
class LineMappingInfo:
    """
    Stores the relationship between physical lines (lines separated by a newline) and
    logical lines (lines connected by line continuation characters or multiline tokens).
    Line numbers are one-indexed.

    Because comments can only exist on or above logical lines (you can't have a comment
    between line continuation characters or inside multiline strings), the linter
    converts physical line numbers to logical line numbers before checking that a
    `# lint-fixme` or `# noqa` comment applies to the line.

    Terminology
    -----------

    Physical Line:
        A line separated by newline characters.
    Logical Line:
        A grouping of physical lines. A logical line always has space for exactly one
        trailing comment.
    Non-Empty Logical Line:
        A logical line that isn't just whitespace or comments. When we encounter a
        `# lint-fixme` comment, we need to map it back to the next non-empty logical
        line, since that's what it likely refers to.
    """

    # A many-to-one mapping (1-indexed)
    physical_to_logical: Mapping[int, int]
    # A sorted list of logical lines that have "real" code on them
    non_empty_logical_lines: Sequence[int]

    def get_next_non_empty_logical_line(self, starting_line: int) -> Optional[int]:
        haystack = self.non_empty_logical_lines
        idx = bisect_left(haystack, starting_line)
        if idx == len(haystack):
            return None
        return haystack[idx]

    @staticmethod
    def compute(*, tokens: Iterable[tokenize.TokenInfo]) -> "LineMappingInfo":
        physical_to_logical = {}
        non_empty_logical_lines = []
        has_content = False
        logical_line_start = 1

        for tok in tokens:
            tok_line = tok.start[0]

            if tok.type in _LOGICAL_LINE_END_MARKERS:
                logical_line_end = tok_line
                physical_lines = range(logical_line_start, logical_line_end + 1)
                for pl in physical_lines:
                    physical_to_logical[pl] = logical_line_start
                if has_content or tok.type == tokenize.ENDMARKER:
                    non_empty_logical_lines.append(logical_line_start)
                has_content = False
                logical_line_start = logical_line_end + 1
            elif not has_content and tok.type not in _EMPTY_TOKENS:
                has_content = True

        return LineMappingInfo(physical_to_logical, non_empty_logical_lines)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Callable, Union

import libcst as cst
from libcst.metadata import CodePosition, MetadataWrapper
from libcst.testing.utils import data_provider, UnitTest

from fixit.common.autofix import LintPatch


class AutofixTest(UnitTest):
    @data_provider(
        {
            "full_module": {
                "original_module": "# hello, world\ndef foo(): ...\nbar()\n",
                "replacement_module": "val = 1 + 2\nraise Exception()\n",
                "get_original_node": lambda module: module,
                "get_replacement_node": lambda __: cst.parse_module(
                    "val = 1 + 2\nraise Exception()\n"
                ),
            },
            "full_module_noop": {
                "original_module": "# hello, world\ndef foo(): ...\nbar()\n",
                "replacement_module": "# hello, world\ndef foo(): ...\nbar()\n",
                "get_original_node": lambda module: module,
                "get_replacement_node": lambda module: module.deep_clone(),
            },
            "remove_statement": {
                "original_module": "first_line\nsecond_line\n",
                "replacement_module": "second_line\n",
                "get_original_node": lambda module: module.body[0],
                "get_replacement_node": lambda __: cst.RemovalSentinel.REMOVE,
            },
            "first_statement": {
                "original_module": "a\nb\n",
                "replacement_module": "new_statement()\nb\n",
                "get_original_node": lambda module: module.body[0],
                "get_replacement_node": lambda __: cst.parse_statement(
                    "new_statement()"
                ),
            },
            "first_expression": {
                "original_module": "old_fn()\nb\n",
                "replacement_module": "new_fn()\nb\n",
                "get_original_node": lambda module: module.body[0].body[0].value.func,
                "get_replacement_node": lambda __: cst.Name("new_fn"),
            },
            "last_statement": {
                "original_module": "a\nb",
                "replacement_module": "a\nnew_statement()",
                "get_original_node": lambda module: module.body[1],
                "get_replacement_node": lambda __: cst.parse_statement(
                    "new_statement()\n"
                ),
            },
            "last_expression": {
                "original_module": "a\none + two",
                "replacement_module": "a\none + new_value",
                "get_original_node": lambda module: module.body[1].body[0].value.right,
                "get_replacement_node": lambda __: cst.Name("new_value"),
            },
        }
    )
    def test_get(
        self,
        *,
        original_module: str,
        replacement_module: str,
        get_original_node: Callable[[cst.Module], cst.CSTNode],
        get_replacement_node: Callable[
            [cst.CSTNode], Union[cst.CSTNode, cst.RemovalSentinel]
        ],
    ) -> None:
        wrapper = MetadataWrapper(
            cst.parse_module(original_module), unsafe_skip_copy=True
        )
        n = get_original_node(wrapper.module)
        patch = LintPatch.get(wrapper, n, get_replacement_node(n))
        self.assertEqual(patch.apply(original_module), replacement_module)
        self.assertEqual(patch.minimize().apply(original_module), replacement_module)

    @data_provider(
        {
            "non_minimizable": {
                "before": LintPatch(0, CodePosition(1, 0), "foobar", "barfoo"),
                "after": LintPatch(0, CodePosition(1, 0), "foobar", "barfoo"),
            },
            "identical_tail": {
                "before": LintPatch(
                    0, CodePosition(1, 0), "hello, world!\n", "goodbye, world!\n"
                ),
                "after": LintPatch(0, CodePosition(1, 0), "hello", "goodbye"),
            },
            "identical_head": {
                "before": LintPatch(0, CodePosition(1, 0), "who", "what"),
                "after": LintPatch(2, CodePosition(1, 2), "o", "at"),
            },
            "newlines_lf": {
                "before": LintPatch(0, CodePosition(1, 0), "a\nb", "a\nc"),
                "after": LintPatch(2, CodePosition(2, 0), "b", "c"),
            },
            "newlines_cr": {
                "before": LintPatch(0, CodePosition(1, 0), "a\rb", "a\rc"),
                "after": LintPatch(2, CodePosition(2, 0), "b", "c"),
            },
            "newlines_crlf": {
                "before": LintPatch(0, CodePosition(1, 0), "a\r\nb", "a\r\nc"),
                "after": LintPatch(3, CodePosition(2, 0), "b", "c"),
            },
            "newlines_extended": {  # test a mix of multiple newlines in the same file
                "before": LintPatch(
                    0,
                    CodePosition(1, 0),
                    "a\r\nb\nc\rd\r\nis final",
                    "a\r\nb\nc\rd\r\nis last",
                ),
                "after": LintPatch(13, CodePosition(5, 3), "final", "last"),
            },
            "minimizable_noop": {  # should minimize to an empty patch
                "before": LintPatch(
                    0,
                    CodePosition(1, 0),
                    "This is\nsome\ncode\n",
                    "This is\nsome\ncode\n",
                ),
                "after": LintPatch(0, CodePosition(1, 0), "", ""),
            },
        }
    )
    def test_minimize(self, *, before: LintPatch, after: LintPatch) -> None:
        self.assertEqual(before.minimize(), after)
        # this should be a noop
        self.assertEqual(after.minimize(), after)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
For backwards compatibility.
"""

from fixit.cli import find_files, ipc_main, IPCResult, LintOpts, map_paths


__all__ = ["IPCResult", "LintOpts", "find_files", "ipc_main", "map_paths"]



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import libcst as cst
import libcst.matchers as m

from fixit import (
    CstContext,
    CstLintRule,
    InvalidTestCase as Invalid,
    ValidTestCase as Valid,
)


class ReplaceUnionWithOptionalRule(CstLintRule):
    """
    Enforces the use of ``Optional[T]`` over ``Union[T, None]`` and ``Union[None, T]``.
    See https://docs.python.org/3/library/typing.html#typing.Optional to learn more about Optionals.
    """

    MESSAGE: str = (
        "`Optional[T]` is preferred over `Union[T, None]` or `Union[None, T]`. "
        + "Learn more: https://docs.python.org/3/library/typing.html#typing.Optional"
    )
    METADATA_DEPENDENCIES = (cst.metadata.ScopeProvider,)
    VALID = [
        Valid(
            """
            def func() -> Optional[str]:
                pass
            """
        ),
        Valid(
            """
            def func() -> Optional[Dict]:
                pass
            """
        ),
        Valid(
            """
            def func() -> Union[str, int, None]:
                pass
            """
        ),
    ]
    INVALID = [
        Invalid(
            """
            def func() -> Union[str, None]:
                pass
            """,
        ),
        Invalid(
            """
            from typing import Optional
            def func() -> Union[Dict[str, int], None]:
                pass
            """,
            expected_replacement="""
            from typing import Optional
            def func() -> Optional[Dict[str, int]]:
                pass
            """,
        ),
        Invalid(
            """
            from typing import Optional
            def func() -> Union[str, None]:
                pass
            """,
            expected_replacement="""
            from typing import Optional
            def func() -> Optional[str]:
                pass
            """,
        ),
        Invalid(
            """
            from typing import Optional
            def func() -> Union[Dict, None]:
                pass
            """,
            expected_replacement="""
            from typing import Optional
            def func() -> Optional[Dict]:
                pass
            """,
        ),
    ]

    def __init__(self, context: CstContext) -> None:
        super().__init__(context)

    def leave_Annotation(self, original_node: cst.Annotation) -> None:
        if self.contains_union_with_none(original_node):
            scope = self.get_metadata(cst.metadata.ScopeProvider, original_node, None)
            nones = 0
            indexes = []
            replacement = None
            if scope is not None and "Optional" in scope:
                for s in cst.ensure_type(original_node.annotation, cst.Subscript).slice:
                    if m.matches(s, m.SubscriptElement(m.Index(m.Name("None")))):
                        nones += 1
                    else:
                        indexes.append(s.slice)
                if not (nones > 1) and len(indexes) == 1:
                    replacement = original_node.with_changes(
                        annotation=cst.Subscript(
                            value=cst.Name("Optional"),
                            slice=(cst.SubscriptElement(indexes[0]),),
                        )
                    )
                    # TODO(T57106602) refactor lint replacement once extract exists
            self.report(original_node, replacement=replacement)

    def contains_union_with_none(self, node: cst.Annotation) -> bool:
        return m.matches(
            node,
            m.Annotation(
                m.Subscript(
                    value=m.Name("Union"),
                    slice=m.OneOf(
                        [
                            m.SubscriptElement(m.Index()),
                            m.SubscriptElement(m.Index(m.Name("None"))),
                        ],
                        [
                            m.SubscriptElement(m.Index(m.Name("None"))),
                            m.SubscriptElement(m.Index()),
                        ],
                    ),
                )
            ),
        )



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Sequence

import libcst as cst
import libcst.matchers as m
from libcst.helpers import ensure_type

from fixit.common.base import CstLintRule
from fixit.common.utils import InvalidTestCase as Invalid, ValidTestCase as Valid


class UseAssertIsNotNoneRule(CstLintRule):
    """
    Discourages use of ``assertTrue(x is not None)`` and ``assertFalse(x is not None)`` as it is deprecated (https://docs.python.org/3.8/library/unittest.html#deprecated-aliases).
    Use ``assertIsNotNone(x)`` and ``assertIsNone(x)``) instead.

    """

    MESSAGE: str = (
        '"assertTrue" and "assertFalse" are deprecated. Use "assertIsNotNone" and "assertIsNone" instead.\n'
        + "See https://docs.python.org/3.8/library/unittest.html#deprecated-aliases"
    )

    VALID = [
        Valid("self.assertIsNotNone(x)"),
        Valid("self.assertIsNone(x)"),
        Valid("self.assertIsNone(None)"),
        Valid("self.assertIsNotNone(f(x))"),
        Valid("self.assertIsNone(f(x))"),
        Valid("self.assertIsNone(object.key)"),
        Valid("self.assertIsNotNone(object.key)"),
    ]

    INVALID = [
        Invalid(
            "self.assertTrue(a is not None)",
            expected_replacement="self.assertIsNotNone(a)",
        ),
        Invalid(
            "self.assertTrue(not x is None)",
            expected_replacement="self.assertIsNotNone(x)",
        ),
        Invalid(
            "self.assertTrue(f() is not None)",
            expected_replacement="self.assertIsNotNone(f())",
        ),
        Invalid(
            "self.assertTrue(not x is not None)",
            expected_replacement="self.assertIsNone(x)",
        ),
        Invalid(
            "self.assertTrue(f(x) is not None)",
            expected_replacement="self.assertIsNotNone(f(x))",
        ),
        Invalid(
            "self.assertTrue(x is None)", expected_replacement="self.assertIsNone(x)"
        ),
        Invalid(
            "self.assertFalse(x is not None)",
            expected_replacement="self.assertIsNone(x)",
        ),
        Invalid(
            "self.assertFalse(not x is None)",
            expected_replacement="self.assertIsNone(x)",
        ),
        Invalid(
            "self.assertFalse(f() is not None)",
            expected_replacement="self.assertIsNone(f())",
        ),
        Invalid(
            "self.assertFalse(not x is not None)",
            expected_replacement="self.assertIsNotNone(x)",
        ),
        Invalid(
            "self.assertFalse(f(x) is not None)",
            expected_replacement="self.assertIsNone(f(x))",
        ),
        Invalid(
            "self.assertFalse(x is None)",
            expected_replacement="self.assertIsNotNone(x)",
        ),
    ]

    def visit_Call(self, node: cst.Call) -> None:
        match_compare_is_none = m.ComparisonTarget(
            m.SaveMatchedNode(
                m.OneOf(m.Is(), m.IsNot()),
                "comparison_type",
            ),
            comparator=m.Name("None"),
        )
        result = m.extract(
            node,
            m.Call(
                func=m.Attribute(
                    value=m.Name("self"),
                    attr=m.SaveMatchedNode(
                        m.OneOf(m.Name("assertTrue"), m.Name("assertFalse")),
                        "assertion_name",
                    ),
                ),
                args=[
                    m.Arg(
                        m.SaveMatchedNode(
                            m.OneOf(
                                m.Comparison(comparisons=[match_compare_is_none]),
                                m.UnaryOperation(
                                    operator=m.Not(),
                                    expression=m.Comparison(
                                        comparisons=[match_compare_is_none]
                                    ),
                                ),
                            ),
                            "argument",
                        )
                    )
                ],
            ),
        )

        if result:
            assertion_name = result["assertion_name"]
            if isinstance(assertion_name, Sequence):
                assertion_name = assertion_name[0]

            argument = result["argument"]
            if isinstance(argument, Sequence):
                argument = argument[0]

            comparison_type = result["comparison_type"]
            if isinstance(comparison_type, Sequence):
                comparison_type = comparison_type[0]

            if m.matches(argument, m.Comparison()):
                assertion_argument = ensure_type(argument, cst.Comparison).left
            else:
                assertion_argument = ensure_type(
                    ensure_type(argument, cst.UnaryOperation).expression, cst.Comparison
                ).left

            negations_seen = 0
            if m.matches(assertion_name, m.Name("assertFalse")):
                negations_seen += 1
            if m.matches(argument, m.UnaryOperation()):
                negations_seen += 1
            if m.matches(comparison_type, m.IsNot()):
                negations_seen += 1

            new_attr = "assertIsNone" if negations_seen % 2 == 0 else "assertIsNotNone"
            new_call = node.with_changes(
                func=cst.Attribute(value=cst.Name("self"), attr=cst.Name(new_attr)),
                args=[cst.Arg(assertion_argument)],
            )

            if new_call is not node:
                self.report(node, replacement=new_call)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import List, Union

import libcst as cst
from libcst.metadata import (
    Assignment,
    QualifiedName,
    QualifiedNameProvider,
    QualifiedNameSource,
    ScopeProvider,
)

from fixit import CstLintRule, InvalidTestCase as Invalid, ValidTestCase as Valid


CLS = "cls"


class _RenameTransformer(cst.CSTTransformer):
    def __init__(
        self, names: List[Union[cst.Name, cst.BaseString, cst.Attribute]], new_name: str
    ) -> None:
        self.names = names
        self.new_name = new_name

    def leave_Name(self, original_node: cst.Name, updated_node: cst.Name) -> cst.Name:
        # Yes this is linear search. We could potentially create a set out of the list,
        # but in most cases there are so few references to the renamed variable that the
        # overhead of creating a set + computing hashes on lookup would likely outweigh
        # any savings. Did not actually benchmark. This code runs extremely rarely IRL.
        if original_node in self.names:
            return original_node.with_changes(value=self.new_name)
        return updated_node


class UseClsInClassmethodRule(CstLintRule):
    """
    Enforces using ``cls`` as the first argument in a ``@classmethod``.
    """

    METADATA_DEPENDENCIES = (QualifiedNameProvider, ScopeProvider)
    MESSAGE = "When using @classmethod, the first argument must be `cls`."
    VALID = [
        Valid(
            """
            class foo:
                # classmethod with cls first arg.
                @classmethod
                def cm(cls, a, b, c):
                    pass
            """
        ),
        Valid(
            """
            class foo:
                # non-classmethod with non-cls first arg.
                def nm(self, a, b, c):
                    pass
            """
        ),
        Valid(
            """
            class foo:
                # staticmethod with non-cls first arg.
                @staticmethod
                def sm(a):
                    pass
            """
        ),
    ]
    INVALID = [
        Invalid(
            """
            class foo:
                # No args at all.
                @classmethod
                def cm():
                    pass
            """,
            expected_replacement="""
            class foo:
                # No args at all.
                @classmethod
                def cm(cls):
                    pass
            """,
        ),
        Invalid(
            """
            class foo:
                # Single arg + reference.
                @classmethod
                def cm(a):
                    return a
            """,
            expected_replacement="""
            class foo:
                # Single arg + reference.
                @classmethod
                def cm(cls):
                    return cls
            """,
        ),
        Invalid(
            """
            class foo:
                # Another "cls" exists: do not autofix.
                @classmethod
                def cm(a):
                    cls = 2
            """,
        ),
        Invalid(
            """
            class foo:
                # Multiple args + references.
                @classmethod
                async def cm(a, b):
                    b = a
                    b = a.__name__
            """,
            expected_replacement="""
            class foo:
                # Multiple args + references.
                @classmethod
                async def cm(cls, b):
                    b = cls
                    b = cls.__name__
            """,
        ),
        Invalid(
            """
            class foo:
                # Do not replace in nested scopes.
                @classmethod
                async def cm(a, b):
                    b = a
                    b = lambda _: a.__name__
                    def g():
                        return a.__name__

                    # Same-named vars in sub-scopes should not be replaced.
                    b = [a for a in [1,2,3]]
                    def f(a):
                        return a + 1
            """,
            expected_replacement="""
            class foo:
                # Do not replace in nested scopes.
                @classmethod
                async def cm(cls, b):
                    b = cls
                    b = lambda _: cls.__name__
                    def g():
                        return cls.__name__

                    # Same-named vars in sub-scopes should not be replaced.
                    b = [a for a in [1,2,3]]
                    def f(a):
                        return a + 1
            """,
        ),
        Invalid(
            """
            # Do not replace in surrounding scopes.
            a = 1

            class foo:
                a = 2

                def im(a):
                    a = a

                @classmethod
                def cm(a):
                    a[1] = foo.cm(a=a)
            """,
            expected_replacement="""
            # Do not replace in surrounding scopes.
            a = 1

            class foo:
                a = 2

                def im(a):
                    a = a

                @classmethod
                def cm(cls):
                    cls[1] = foo.cm(a=cls)
            """,
        ),
        Invalid(
            """
            def another_decorator(x): pass

            class foo:
                # Multiple decorators.
                @another_decorator
                @classmethod
                @another_decorator
                async def cm(a, b, c):
                    pass
            """,
            expected_replacement="""
            def another_decorator(x): pass

            class foo:
                # Multiple decorators.
                @another_decorator
                @classmethod
                @another_decorator
                async def cm(cls, b, c):
                    pass
            """,
        ),
    ]

    def visit_FunctionDef(self, node: cst.FunctionDef) -> None:
        if not any(
            QualifiedNameProvider.has_name(
                self,
                decorator.decorator,
                QualifiedName(
                    name="builtins.classmethod", source=QualifiedNameSource.BUILTIN
                ),
            )
            for decorator in node.decorators
        ):
            return  # If it's not a @classmethod, we are not interested.
        if not node.params.params:
            # No params, but there must be the 'cls' param.
            # Note that pyre[47] already catches this, but we also generate
            # an autofix, so it still makes sense for us to report it here.
            new_params = node.params.with_changes(
                params=(cst.Param(name=cst.Name(value=CLS)),)
            )
            repl = node.with_changes(params=new_params)
            self.report(node, replacement=repl)
            return

        p0_name = node.params.params[0].name
        if p0_name.value == CLS:
            return  # All good.

        # Rename all assignments and references of the first param within the
        # function scope, as long as they are done via a Name node.
        # We rely on the parser to correctly derive all
        # assigments and references within the FunctionScope.
        # The Param node's scope is our classmethod's FunctionScope.
        scope = self.get_metadata(ScopeProvider, p0_name, None)
        if not scope:
            # Cannot autofix without scope metadata. Only report in this case.
            # Not sure how to repro+cover this in a unit test...
            # If metadata creation fails, then the whole lint fails, and if it succeeds,
            # then there is valid metadata. But many other lint rule implementations contain
            # a defensive scope None check like this one, so I assume it is necessary.
            self.report(node)
            return

        if scope[CLS]:
            # The scope already has another assignment to "cls".
            # Trying to rename the first param to "cls" as well may produce broken code.
            # We should therefore refrain from suggesting an autofix in this case.
            self.report(node)
            return

        refs: List[Union[cst.Name, cst.Attribute, cst.BaseString]] = []
        assignments = scope[p0_name.value]
        for a in assignments:
            if isinstance(a, Assignment):
                assign_node = a.node
                if isinstance(assign_node, cst.Name):
                    refs.append(assign_node)
                elif isinstance(assign_node, cst.Param):
                    refs.append(assign_node.name)
                # There are other types of possible assignment nodes: ClassDef,
                # FunctionDef, Import, etc. We deliberately do not handle those here.
            refs += [r.node for r in a.references]

        repl = node.visit(_RenameTransformer(refs, CLS))
        self.report(node, replacement=repl)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import libcst as cst
import libcst.matchers as m

from fixit import CstLintRule, InvalidTestCase as Invalid, ValidTestCase as Valid


class NoInheritFromObjectRule(CstLintRule):
    """
    In Python 3, a class is inherited from ``object`` by default.
    Explicitly inheriting from ``object`` is redundant, so removing it keeps the code simpler.
    """

    MESSAGE = "Inheriting from object is a no-op.  'class Foo:' is just fine =)"
    VALID = [
        Valid("class A(something):    pass"),
        Valid(
            """
            class A:
                pass"""
        ),
    ]
    INVALID = [
        Invalid(
            """
            class B(object):
                pass""",
            line=1,
            column=1,
            expected_replacement="""
                class B:
                    pass""",
        ),
        Invalid(
            """
            class B(object, A):
                pass""",
            line=1,
            column=1,
            expected_replacement="""
                class B(A):
                    pass""",
        ),
    ]

    def visit_ClassDef(self, node: cst.ClassDef) -> None:
        new_bases = tuple(
            base for base in node.bases if not m.matches(base.value, m.Name("object"))
        )

        if tuple(node.bases) != new_bases:
            # reconstruct classdef, removing parens if bases and keywords are empty
            new_classdef = node.with_changes(
                bases=new_bases,
                lpar=cst.MaybeSentinel.DEFAULT,
                rpar=cst.MaybeSentinel.DEFAULT,
            )

            # report warning and autofix
            self.report(node, replacement=new_classdef)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
import shutil
from pathlib import Path

from libcst.testing.utils import UnitTest

from fixit.common.config import get_lint_config, get_rules_from_config
from fixit.common.utils import (
    dedent_with_lstrip,
    DuplicateLintRuleNameError,
    find_and_import_rule,
    import_rule_from_package,
    LintRuleNotFoundError,
)


DUMMY_PACKAGE: str = "fixit.common.tests.dummy_package"
DUMMY_SUBPACKAGE: str = "fixit.common.tests.dummy_package.dummy_subpackage"
DUMMY_CONFIG_PATH: str = ".fixit.config.yaml"

# Using dummy config file, test whether the rule import helpers work as expected.
class ImportsTest(UnitTest):
    def setUp(self) -> None:
        # We need to change the working directory so that the dummy config file is used.
        self.old_wd = os.getcwd()
        test_dir = Path(__file__).parent
        os.chdir(test_dir)

        # We also need to clear the lru_cache for the get_lint_config function between tests.
        getattr(get_lint_config, "cache_clear")()

    def tearDown(self) -> None:
        # Need to change back to original working directory so that we don't mess with other unit tests.
        os.chdir(self.old_wd)
        getattr(get_lint_config, "cache_clear")()

    def test_get_rules_from_config(self) -> None:
        with self.assertRaises(DuplicateLintRuleNameError):
            # We have two dummy lint rules with the same name. Verify this raises an error.
            get_rules_from_config()

        # Now try to import all rules from a package where there aren't any duplicates.
        getattr(get_lint_config, "cache_clear")()
        next_dir = os.path.join(os.getcwd(), "dummy_package")
        os.chdir(next_dir)
        rules = get_rules_from_config()

        # Assert all rules are imported as expected.
        self.assertEqual(len(rules), 3)
        self.assertTrue(all(r.__module__ == f"{DUMMY_SUBPACKAGE}.dummy" for r in rules))

    def test_import_rule_from_package(self) -> None:
        rules_package = get_lint_config().packages
        self.assertEqual(rules_package, [DUMMY_PACKAGE])

        # Test with an existing dummy rule.
        imported_rule = import_rule_from_package(rules_package[0], "DummyRule2")
        self.assertIsNotNone(imported_rule)
        self.assertEqual(imported_rule.__name__, "DummyRule2")
        self.assertEqual(imported_rule.__module__, f"{DUMMY_PACKAGE}.dummy_2")

        # Test with non-existent rule.
        imported_rule = import_rule_from_package(rules_package[0], "DummyRule1000")
        self.assertIsNone(imported_rule)

    def test_find_and_import_rule(self) -> None:
        rules_packages = get_lint_config().packages

        # Test with existing dummy rule. Should get the first one it finds, from dummy_1 module.
        imported_rule = find_and_import_rule("DummyRule1", rules_packages)
        self.assertEqual(imported_rule.__module__, f"{DUMMY_PACKAGE}.dummy_1")

        with self.assertRaises(LintRuleNotFoundError):
            imported_rule = find_and_import_rule("DummyRule1000", rules_packages)


class AllowListTest(UnitTest):
    def setUp(self) -> None:
        # We need to change the working directory so that the dummy config file is used.
        self.old_wd = os.getcwd()
        test_dir = Path(__file__).parent
        os.chdir(test_dir)
        next_dir = os.path.join(os.getcwd(), "dummy_package")
        os.chdir(next_dir)

        # We also need to clear the lru_cache for the get_lint_config function between tests.
        getattr(get_lint_config, "cache_clear")()
        shutil.copyfile(DUMMY_CONFIG_PATH, "test.fixit.config.yaml")

    def tearDown(self) -> None:
        shutil.copyfile("test.fixit.config.yaml", DUMMY_CONFIG_PATH)
        os.remove("test.fixit.config.yaml")
        # Need to change back to original working directory so that we don't mess with other unit tests.
        os.chdir(self.old_wd)
        getattr(get_lint_config, "cache_clear")()

    def test_allow_list_rules_omit_one(self) -> None:
        allow_block = dedent_with_lstrip(
            """
            allow_list_rules:
            - DummyRule1
            - DummyRule2
            """
        )
        with open(DUMMY_CONFIG_PATH, "a") as f:
            f.write(allow_block)
        rules = get_rules_from_config()

        # Only the two rules listed as imported
        self.assertEqual(len(rules), 2)
        self.assertTrue(all(r.__module__ == f"{DUMMY_SUBPACKAGE}.dummy" for r in rules))

    def test_allow_list_rules_block_list_overrides(self) -> None:
        allow_block = dedent_with_lstrip(
            """
            allow_list_rules:
            - DummyRule1
            - DummyRule2
            - DummyRule4
            """
        )
        with open(DUMMY_CONFIG_PATH, "a") as f:
            f.write(allow_block)
        rules = get_rules_from_config()

        # DummyRule4 is still not included because it is in the block_rules_list
        self.assertEqual(len(rules), 2)
        self.assertTrue(all(r.__module__ == f"{DUMMY_SUBPACKAGE}.dummy" for r in rules))

    def test_allow_list_rules_empty(self) -> None:
        allow_block = "allow_list_rules: []"
        with open(DUMMY_CONFIG_PATH, "a") as f:
            f.write(allow_block)
        rules = get_rules_from_config()
        # All rules should be imported.
        self.assertEqual(len(rules), 3)
        self.assertTrue(all(r.__module__ == f"{DUMMY_SUBPACKAGE}.dummy" for r in rules))



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from fixit.common.base import CstContext, CstLintRule, LintConfig
from fixit.common.report import CstLintRuleReport
from fixit.common.testing import add_lint_rule_tests_to_module
from fixit.common.utils import InvalidTestCase, ValidTestCase


__all__ = [
    "CstContext",
    "CstLintRule",
    "LintConfig",
    "ValidTestCase",
    "InvalidTestCase",
    "CstLintRuleReport",
    "add_lint_rule_tests_to_module",
]



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import tokenize
from dataclasses import dataclass
from typing import Iterable, Sequence


@dataclass(frozen=True)
class CommentInfo:

    # All comment tokens. A superset of `comments_on_own_line`. These may be trailing
    # comments, meaning that they come after other non-whitespace content on the same
    # line.
    comments: Sequence[tokenize.TokenInfo]

    # A comment on a line with no other leading tokens is a "comment on own line". This
    # is a subset of all of the comments.
    comments_on_own_line: Sequence[tokenize.TokenInfo]

    @staticmethod
    def compute(*, tokens: Iterable[tokenize.TokenInfo]) -> "CommentInfo":
        comments = []
        comments_on_own_line = []
        prev_tok = None
        for tok in tokens:
            if tok.type == tokenize.COMMENT:
                comments.append(tok)
                # Find comments on a line by themselves
                # https://gitlab.com/pycqa/flake8/merge_requests/219
                #
                # We need to check the previous token, but there's no need to check for
                # the next token, since a comment is always the last thing on a given
                # line.
                if prev_tok is None or prev_tok.end[0] != tok.start[0]:
                    comments_on_own_line.append(tok)
            prev_tok = tok
        return CommentInfo(comments, comments_on_own_line)



# Copyright (c) Meta Platforms, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Linting framework built on LibCST, with automatic fixes
"""

__version__ = "0.2.0"

from .api import fixit_bytes, fixit_file, fixit_paths
from .types import Config, FileContent, Result

__all__ = [
    "__version__",
]



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# Usage:
#
#   $ fixit run_rules --help
#   $ fixit apply_fix --help
#   $ fixit insert_suppressions --help
#   $ fixit add_new_rule --help

import argparse
import sys
from typing import List

from fixit._version import FIXIT_VERSION
from fixit.cli import add_new_rule, apply_fix, insert_suppressions, run_rules


def main(argv: List[str] = sys.argv[1:]) -> None:
    """Fixit CLI and its sub-commands"""
    parser = argparse.ArgumentParser(
        prog="fixit",
        description="These are common Fixit commands used in various situations:",
    )

    parser.add_argument("--version", action="version", version=FIXIT_VERSION)
    parser.set_defaults(subparser_fn=lambda _: parser.print_help())

    # Register sub-commands
    subparser = parser.add_subparsers(title="command")

    add_new_rule.register_subparser(subparser)
    apply_fix.register_subparser(subparser)
    insert_suppressions.register_subparser(subparser)
    run_rules.register_subparser(subparser)

    args = parser.parse_args(argv)
    sys.exit(args.subparser_fn(args))


if __name__ == "__main__":
    main()



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import argparse
import ast
import io
import tokenize
from functools import lru_cache
from typing import Any, List, Optional, Type

from flake8.checker import (
    FileChecker as Flake8FileChecker,
    Manager as Flake8CheckerManager,
)
from flake8.formatting.base import BaseFormatter as Flake8BaseFormatter
from flake8.main.application import Application as Flake8BaseApplication
from flake8.processor import FileProcessor as Flake8FileProcessor
from flake8.style_guide import Violation as Flake8Violation

from fixit.common.pseudo_rule import PseudoContext
from fixit.common.report import BaseLintRuleReport


__all__ = ["Flake8LintRuleReport"]


class Flake8CompatAccumulatingFormatter(Flake8BaseFormatter):
    """
    Handles errors by writing them to the provided `accumulator` object. Does not write
    anything to stdout/stderr/a file.
    """

    def __init__(
        self, options: argparse.Namespace, accumulator: List[Flake8Violation]
    ) -> None:
        super().__init__(options)
        self.accumulator: List[Flake8Violation] = accumulator

    def handle(self, error: Flake8Violation) -> None:
        self.accumulator.append(error)

    def start(self) -> None:
        # By default, this may open a file. Don't do that!
        pass

    def stop(self) -> None:
        # By default, this may close a file. Don't do that!
        pass

    def write(self, line: Optional[str], source: Optional[str]) -> None:
        # Responsible for writing output to stdout/stderr/etc. Don't do that!
        pass


class Flake8CompatFileProcessor(Flake8FileProcessor):
    def __init__(self, *args: Any, context: PseudoContext, **kwargs: Any) -> None:
        self.context: PseudoContext = (
            context  # super calls read_lines; do this before super()
        )
        super().__init__(*args, **kwargs)

    @property
    def file_tokens(self) -> List[tokenize.TokenInfo]:
        tokens_iter = iter(self.context.tokens)
        first_token = next(tokens_iter)  # strip the leading ENCODING token
        assert first_token.type == tokenize.ENCODING
        return list(tokens_iter)

    def build_ast(self) -> ast.Module:
        return self.context.ast_tree

    def read_lines(self) -> List[str]:
        encoding, __ = tokenize.detect_encoding(
            io.BytesIO(self.context.source).readline
        )
        decoded = self.context.source.decode(encoding)
        return io.StringIO(decoded).readlines()


class Flake8CompatFileChecker(Flake8FileChecker):
    def __init__(self, *args: Any, context: PseudoContext, **kwargs: Any) -> None:
        self.context = context  # super calls _make_processor; do this before super()
        super().__init__(*args, **kwargs)

    def _make_processor(self) -> Flake8CompatFileProcessor:
        return Flake8CompatFileProcessor(
            self.filename, self.options, context=self.context
        )


class Flake8CompatCheckerManager(Flake8CheckerManager):
    def __init__(self, *args: Any, context: PseudoContext, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.context = context

    def run(self) -> None:
        # Normally this will detect the number of jobs and use multiprocessing if it's
        # supported by the interpreter.
        #
        # Our lint framework does its own parallelization, so we want to disable that
        # here.
        self.run_serial()

    def make_checkers(self, paths: Optional[List[str]] = None) -> None:
        assert paths is not None
        assert len(paths) == 1

        checks = self.checks.to_dictionary()
        checkers = [
            Flake8CompatFileChecker(
                paths[0], checks, self.options, context=self.context
            )
        ]
        # Necessary for the flake8 changes introduced here https://github.com/PyCQA/flake8/commit/bfb79b46c807168dbc25fd1e9e41359c4558256f
        # Flake8's Manager class now uses the parameter self._all_checkers to store the checkers
        self._all_checkers = self.checkers = [
            checker for checker in checkers if checker.should_process
        ]


class Flake8CompatApplication(Flake8BaseApplication):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.accumulator: List[Flake8Violation] = []

    def reset(self, context: PseudoContext) -> None:
        # An Application isn't really intended to be used multiple times, but we want to
        # avoid re-instantiating it because of the performance overhead, so let's just
        # reset the file checker manager. The file checker manager is the thing that
        # actually holds the reports generated by checkers.
        self.file_checker_manager = Flake8CompatCheckerManager(
            style_guide=self.guide,
            arguments=self.args,
            checker_plugins=self.check_plugins,
            context=context,
        )
        # And clear the list that Flake8CompatAccumulatingFormatter writes to
        self.accumulator.clear()

    def make_formatter(
        self, formatter_class: Optional[Type[Flake8BaseFormatter]] = None
    ) -> None:
        options = self.options
        assert options is not None
        # ignores formatter_class if given, we don't support flake8 formatters
        self.formatter = Flake8CompatAccumulatingFormatter(options, self.accumulator)

    def report_statistics(self) -> None:
        pass

    def report_benchmarks(self) -> None:
        pass


class Flake8LintRuleReport(BaseLintRuleReport):
    pass


@lru_cache(maxsize=1)
def get_cached_application_instance() -> Flake8CompatApplication:
    """
    The application object is somewhat expensive to construct (it reads flake8
    configuration files), so there shouldn't be more than one instance per worker
    process.
    """
    application = Flake8CompatApplication()
    # disable noqa because we have our own ignore system
    application.initialize(["--disable-noqa"])
    return application



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import argparse
import json
import tempfile
from pathlib import Path
from typing import cast, List

from libcst.metadata.type_inference_provider import (
    _process_pyre_data,
    PyreData,
    run_command,
)

from fixit.cli.args import (
    get_pyre_fixture_dir_parser,
    get_rule_parser,
    get_rules_package_parser,
)
from fixit.common.base import CstLintRule, LintRuleT
from fixit.common.utils import _dedent


class RuleNotTypeDependentError(Exception):
    pass


class RuleTypeError(Exception):
    pass


class PyreQueryError(Exception):
    def __init__(self, command: List[str], message: str) -> None:
        super().__init__(
            "Unable to infer types from temporary file. "
            + f"Command `{' '.join(command)}` returned with the following message: {message}."
        )


def gen_types_for_test_case(source_code: str, dest_path: Path) -> None:
    rule_fixture_subdir: Path = dest_path.parent
    if not rule_fixture_subdir.exists():
        rule_fixture_subdir.mkdir(parents=True)
    with tempfile.NamedTemporaryFile(
        "w", dir=rule_fixture_subdir, suffix=".py"
    ) as temp:
        temp.write(_dedent(source_code))
        temp.seek(0)

        cmd = ["pyre", "query", f'''"types(path='{temp.name}')"''']
        stdout, stderr, return_code = run_command(cmd)
        if return_code != 0:
            raise PyreQueryError(cmd, f"{stdout}\n{stderr}")
        json_data = json.loads(stdout)
        # Check if error is a key in `data` since pyre may report errors this way.
        if "error" in json_data:
            raise PyreQueryError(cmd, json_data["error"])
        data: PyreData = _process_pyre_data(json_data["response"][0])
        print(f"Writing output to {dest_path}")
        dest_path.write_text(json.dumps({"types": data["types"]}, indent=2))


def gen_types(rule: CstLintRule, rule_fixture_dir: Path) -> None:
    if not rule.requires_metadata_caches():
        raise RuleNotTypeDependentError(
            "Rule does not list any cache-dependent providers in its `METADATA_DEPENDENCIES`."
        )
    if hasattr(rule, "VALID") or hasattr(rule, "INVALID"):
        print("Starting pyre server")

        stdout, stderr, return_code = run_command(["pyre", "start"])
        if return_code != 0:
            print(stdout)
            print(stderr)
        else:
            # pyre-fixme[16]: `CstLintRule` has no attribute `__name__`.
            class_name = getattr(rule, "__name__")
            if hasattr(rule, "VALID"):
                # pyre-fixme[16]: `CstLintRule` has no attribute `VALID`.
                for idx, valid_tc in enumerate(getattr(rule, "VALID")):
                    path: Path = rule_fixture_dir / f"{class_name}_VALID_{idx}.json"
                    gen_types_for_test_case(source_code=valid_tc.code, dest_path=path)
            if hasattr(rule, "INVALID"):
                # pyre-fixme[16]: `CstLintRule` has no attribute `INVALID`.
                for idx, invalid_tc in enumerate(getattr(rule, "INVALID")):
                    path: Path = rule_fixture_dir / f"{class_name}_INVALID_{idx}.json"
                    gen_types_for_test_case(source_code=invalid_tc.code, dest_path=path)
            run_command(["pyre", "stop"])


def get_fixture_path(
    fixture_top_dir: Path, rule_module: str, rules_package: str
) -> Path:
    subpackage: str = rule_module.split(f"{rules_package}.", 1)[-1]
    fixture_subdir = subpackage.replace(".", "/")
    return fixture_top_dir / fixture_subdir


if __name__ == "__main__":
    """
    Run this script directly to generate pyre data for a lint rule that requires TypeInferenceProvider metadata.
    """
    parser = argparse.ArgumentParser(
        description="Generate fixture files required to run unit tests on `TypeInference`-dependent lint rules.",
        parents=[
            get_rule_parser(),
            get_pyre_fixture_dir_parser(),
            get_rules_package_parser(),
        ],
    )
    args: argparse.Namespace = parser.parse_args()
    rule: LintRuleT = args.rule
    fixture_dir: Path = args.fixture_dir
    fixture_path: Path = get_fixture_path(
        fixture_dir, rule.__module__, args.rules_package
    )
    if not issubclass(rule, CstLintRule):
        raise RuleTypeError("Rule must inherit from CstLintRule.")
    gen_types(cast(CstLintRule, rule), fixture_path)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import List

import libcst as cst
import libcst.matchers as m

from fixit import (
    CstContext,
    CstLintRule,
    InvalidTestCase as Invalid,
    LintConfig,
    ValidTestCase as Valid,
)


class AddMissingHeaderRule(CstLintRule):
    """
    Verify if required header comments exist in a module.
    Configuration is required in ``.fixit.config.yaml`` in order to enable this rule::

        rule_config:
           AddMissingHeaderRule:
               path: pkg/*.py
               header: |-
                   # header line 1
                   # header line 2

    (Use ``|-`` to keep newlines and add no new line at the end of header comments.)
    ``path`` is a glob-style ``str`` used in `Path.match() <https://docs.python.org/3/library/pathlib.html#pathlib.PurePath.match>`_.
    The specified ``header`` is a newline-separated ``str`` to be enforced in files whose path matches.
    """

    MESSAGE: str = "A required header comment for this file is missing."

    VALID = [
        Valid("import libcst"),
        Valid(
            """
            # header line 1
            # header line 2
            import libcst
            """,
            config=LintConfig(
                rule_config={
                    "AddMissingHeaderRule": {
                        "path": "*.py",
                        "header": "# header line 1\n# header line 2",
                    }
                }
            ),
        ),
        Valid(
            """
            # header line 1
            # header line 2
            # An extra header line is ok
            import libcst
            """,
            config=LintConfig(
                rule_config={
                    "AddMissingHeaderRule": {
                        "path": "*.py",
                        "header": "# header line 1\n# header line 2",
                    }
                }
            ),
        ),
        Valid(
            """
            # other header in an unrelated file
            import libcst
            """,
            filename="b/m.py",
            config=LintConfig(
                rule_config={
                    "AddMissingHeaderRule": {
                        "path": "a/*.py",
                        "header": "# header line 1\n# header line 2",
                    }
                }
            ),
        ),
    ]
    INVALID = [
        Invalid(
            "# wrong header",
            config=LintConfig(
                rule_config={
                    "AddMissingHeaderRule": {"path": "*.py", "header": "# header line"}
                }
            ),
            expected_replacement="# header line\n# wrong header",
        ),
        Invalid(
            """
            #!/usr/bin/env python
            # wrong header""",
            config=LintConfig(
                rule_config={
                    "AddMissingHeaderRule": {"path": "*.py", "header": "# header line"}
                }
            ),
            expected_replacement="#!/usr/bin/env python\n# header line\n# wrong header",
        ),
    ]

    def __init__(self, context: CstContext) -> None:
        super().__init__(context)
        config = self.context.config.rule_config.get(self.__class__.__name__, None)
        if config is None:
            self.rule_disabled: bool = True
        else:
            if not isinstance(config, dict) or "header" not in config:
                raise ValueError(
                    "A ``header`` str config is required by AddMissingHeaderRule."
                )
            header_str = config["header"]
            if not isinstance(header_str, str):
                raise ValueError(
                    "A ``header`` str config is required by AddMissingHeaderRule."
                )
            lines = header_str.split("\n")
            self.header_matcher: List[m.EmptyLine] = [
                m.EmptyLine(comment=m.Comment(value=line)) for line in lines
            ]
            self.header_replacement: List[cst.EmptyLine] = [
                cst.EmptyLine(comment=cst.Comment(value=line)) for line in lines
            ]
            if "path" in config:
                path_pattern = config["path"]
                if not isinstance(path_pattern, str):
                    raise ValueError(
                        "``path`` config should be a str in AddMissingHeaderRule."
                    )
            else:
                path_pattern = "*.py"
            self.rule_disabled = not self.context.file_path.match(path_pattern)

    def visit_Module(self, node: cst.Module) -> None:
        if self.rule_disabled:
            return
        if not m.matches(node, m.Module(header=[*self.header_matcher, m.ZeroOrMore()])):
            shebang, header = [], node.header
            if header:
                comment = header[0].comment
                if isinstance(comment, cst.Comment) and comment.value.startswith("#!"):
                    shebang, header = [header[0]], header[1:]
            self.report(
                node,
                replacement=node.with_changes(
                    header=[*shebang, *self.header_replacement, *header]
                ),
            )



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
For backwards compatibility.
"""

from fixit.cli.args import (
    get_compact_parser,
    get_multiprocessing_parser,
    get_paths_parser,
    get_rule_parser,
    get_skip_ignore_byte_marker_parser,
    get_use_ignore_comments_parser,
)


__all__ = [
    "get_compact_parser",
    "get_multiprocessing_parser",
    "get_paths_parser",
    "get_rule_parser",
    "get_skip_ignore_byte_marker_parser",
    "get_use_ignore_comments_parser",
]



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import re
from typing import Callable, cast, List, Optional

import libcst as cst
import libcst.matchers as m

from fixit.rule.cst import CstLintRule


def Invalid(*args, **kwargs) -> str:
    return "TODO"


def Valid(*args, **kwargs) -> str:
    return "TODO"


USE_FSTRING_SIMPLE_EXPRESSION_MAX_LENGTH = 30


def _match_simple_string(node: cst.BaseExpression) -> bool:
    if isinstance(node, cst.SimpleString) and "b" not in node.prefix.lower():
        # SimpleString can be a bytes and fstring don't support bytes
        # This helper is for autofixer and it only handles %s simple cases for now.
        # for the other types like %f %d, it require more processing on the format.
        # Convert a %d to fstring will change the exception type. We may get help
        # from TypeInferenceProvider to confirm the type before codedmod.
        # We leave it as future work.
        return re.fullmatch("[^%]*(%s[^%]*)+", node.raw_value) is not None
    return False


def _gen_match_simple_expression(
    codegen: Callable[[cst.CSTNode], str],
) -> Callable[[cst.BaseExpression], bool]:
    def _match_simple_expression(node: cst.BaseExpression) -> bool:
        # either each element in Tuple is simple expression or the entire expression is simple.
        if (
            isinstance(node, cst.Tuple)
            and all(
                len(codegen(elm.value)) < USE_FSTRING_SIMPLE_EXPRESSION_MAX_LENGTH
                for elm in node.elements
            )
        ) or len(codegen(node)) < USE_FSTRING_SIMPLE_EXPRESSION_MAX_LENGTH:
            return True
        return False

    return _match_simple_expression


class EscapeStringQuote(cst.CSTTransformer):
    def __init__(self, quote: str) -> None:
        self.quote = quote
        super().__init__()

    def leave_SimpleString(
        self, original_node: cst.SimpleString, updated_node: cst.SimpleString
    ) -> cst.SimpleString:
        if self.quote == original_node.quote:
            for quo in ["'", '"', "'''", '"""']:
                if quo != original_node.quote and quo not in original_node.raw_value:
                    escaped_string = cst.SimpleString(
                        original_node.prefix + quo + original_node.raw_value + quo
                    )
                    if escaped_string.evaluated_value != original_node.evaluated_value:
                        raise Exception(
                            f"Failed to escape string:\n  original:{original_node.value}\n  escaped:{escaped_string.value}"
                        )
                    else:
                        return escaped_string
            raise Exception(
                f"Cannot find a good quote for escaping the SimpleString: {original_node.value}"
            )
        return original_node


class UseFstringRule(CstLintRule):
    """
    Encourages the use of f-string instead of %-formatting or .format() for high code quality and efficiency.

    Following two cases not covered:

    1. arguments length greater than 30 characters: for better readibility reason
        For example:

        1: this is the answer: %d" % (a_long_function_call() + b_another_long_function_call())
        2: f"this is the answer: {a_long_function_call() + b_another_long_function_call()}"
        3: result = a_long_function_call() + b_another_long_function_call()
        f"this is the answer: {result}"

        Line 1 is more readable than line 2. Ideally, we’d like developers to manually fix this case to line 3

    2. only %s placeholders are linted against for now. We leave it as future work to support other placeholders.
        For example, %d raises TypeError for non-numeric objects, whereas f“{x:d}” raises ValueError.
        This discrepancy in the type of exception raised could potentially break the logic in the code where the exception is handled
    """

    MESSAGE: str = (
        "Do not use printf style formatting or .format(). "
        + "Use f-string instead to be more readable and efficient. "
        + "See https://www.python.org/dev/peps/pep-0498/"
    )

    VALID = [
        Valid("somebody='you'; f\"Hey, {somebody}.\""),
        Valid('"hey"'),
        Valid('"hey" + "there"'),
        Valid('b"a type %s" % var'),
    ]

    INVALID = [
        Invalid('"Hey, {somebody}.".format(somebody="you")'),
        Invalid('"%s" % "hi"', expected_replacement='''f"{'hi'}"'''),
        Invalid('"a name: %s" % name', expected_replacement='f"a name: {name}"'),
        Invalid(
            '"an attribute %s ." % obj.attr',
            expected_replacement='f"an attribute {obj.attr} ."',
        ),
        Invalid(
            'r"raw string value=%s" % val',
            expected_replacement='fr"raw string value={val}"',
        ),
        Invalid('"{%s}" % val', expected_replacement='f"{{{val}}}"'),
        Invalid('"{%s" % val', expected_replacement='f"{{{val}"'),
        Invalid(
            '"The type of var: %s" % type(var)',
            expected_replacement='f"The type of var: {type(var)}"',
        ),
        Invalid(
            '"%s" % obj.this_is_a_very_long_expression(parameter)["a_very_long_key"]',
        ),
        Invalid(
            '"type of var: %s, value of var: %s" % (type(var), var)',
            expected_replacement='f"type of var: {type(var)}, value of var: {var}"',
        ),
        Invalid(
            "'%s\" double quote is used' % var",
            expected_replacement="f'{var}\" double quote is used'",
        ),
        Invalid(
            '"var1: %s, var2: %s, var3: %s, var4: %s" % (class_object.attribute, dict_lookup["some_key"], some_module.some_function(), var4)',
            expected_replacement='''f"var1: {class_object.attribute}, var2: {dict_lookup['some_key']}, var3: {some_module.some_function()}, var4: {var4}"''',
        ),
        Invalid(
            '"a list: %s" % " ".join(var)',
            expected_replacement='''f"a list: {' '.join(var)}"''',
        ),
    ]

    _codegen: Optional[Callable[[cst.CSTNode], str]]

    def visit_Module(self, node: cst.Module) -> Optional[bool]:
        self._codegen = node.code_for_node
        return super().visit_Module(node)

    def visit_Call(self, node: cst.Call) -> None:
        if m.matches(
            node,
            m.Call(
                func=m.Attribute(value=m.SimpleString(), attr=m.Name(value="format"))
            ),
        ):
            self.report(node)

    def visit_BinaryOperation(self, node: cst.BinaryOperation) -> None:
        codegen = self._codegen
        if not codegen:
            raise ValueError("No codegen found. Have we visited a Module?")

        expr_key = "expr"
        extracts = m.extract(
            node,
            m.BinaryOperation(
                left=m.MatchIfTrue(_match_simple_string),
                operator=m.Modulo(),
                right=m.SaveMatchedNode(
                    m.MatchIfTrue(_gen_match_simple_expression(codegen)),
                    expr_key,
                ),
            ),
        )

        if extracts:
            expr = cast(cst.BaseExpression, extracts[expr_key])
            parts: List[cst.BaseFormattedStringContent] = []
            simple_string = cst.ensure_type(node.left, cst.SimpleString)
            innards = simple_string.raw_value.replace("{", "{{").replace("}", "}}")
            tokens = innards.split("%s")
            token = tokens[0]
            if len(token) > 0:
                parts.append(cst.FormattedStringText(value=token))
            expressions = (
                [elm.value for elm in expr.elements]
                if isinstance(expr, cst.Tuple)
                else [expr]
            )
            escape_transformer = EscapeStringQuote(simple_string.quote)
            i = 1
            while i < len(tokens):
                if i - 1 >= len(expressions):
                    # Only generate warning for cases where %-string not comes with same number of elements in tuple
                    self.report(node)
                    return
                try:
                    parts.append(
                        cst.FormattedStringExpression(
                            expression=cast(
                                cst.BaseExpression,
                                # pyre-fixme[16]: `Sequence` has no attribute `visit`.
                                expressions[i - 1].visit(escape_transformer),
                            )
                        )
                    )
                except Exception:
                    self.report(node)
                    return
                token = tokens[i]
                if len(token) > 0:
                    parts.append(cst.FormattedStringText(value=token))
                i += 1
            start = f"f{simple_string.prefix}{simple_string.quote}"
            replacement = cst.FormattedString(
                parts=parts, start=start, end=simple_string.quote
            )
            self.report(node, replacement=replacement)
        elif m.matches(
            node, m.BinaryOperation(left=m.SimpleString(), operator=m.Modulo())
        ) and isinstance(
            cst.ensure_type(node.left, cst.SimpleString).evaluated_value, str
        ):
            self.report(node)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import List, Optional, Sequence, Tuple

import libcst as cst
from libcst import ensure_type, MaybeSentinel, parse_expression
from libcst.metadata import QualifiedName, QualifiedNameProvider, QualifiedNameSource

from fixit import CstLintRule, InvalidTestCase as Invalid, ValidTestCase as Valid


class NoNamedTupleRule(CstLintRule):
    """
    Enforce the use of ``dataclasses.dataclass`` decorator instead of ``NamedTuple`` for cleaner customization and
    inheritance. It supports default value, combining fields for inheritance, and omitting optional fields at
    instantiation. See `PEP 557 <https://www.python.org/dev/peps/pep-0557>`_.
    ``@dataclass`` is faster at reading an object's nested properties and executing its methods. (`benchmark <https://medium.com/@jacktator/dataclass-vs-namedtuple-vs-object-for-performance-optimization-in-python-691e234253b9>`_)
    """

    MESSAGE: str = "Instead of NamedTuple, consider using the @dataclass decorator from dataclasses instead for simplicity, efficiency and consistency."
    METADATA_DEPENDENCIES = (QualifiedNameProvider,)

    VALID = [
        Valid(
            """
            @dataclass(frozen=True)
            class Foo:
                pass
            """
        ),
        Valid(
            """
            @dataclass(frozen=False)
            class Foo:
                pass
            """
        ),
        Valid(
            """
            class Foo:
                pass
            """
        ),
        Valid(
            """
            class Foo(SomeOtherBase):
                pass
            """
        ),
        Valid(
            """
            @some_other_decorator
            class Foo:
                pass
            """
        ),
        Valid(
            """
            @some_other_decorator
            class Foo(SomeOtherBase):
                pass
            """
        ),
    ]
    INVALID = [
        Invalid(
            code="""
            from typing import NamedTuple

            class Foo(NamedTuple):
                pass
            """,
            expected_replacement="""
            from typing import NamedTuple

            @dataclass(frozen=True)
            class Foo:
                pass
            """,
        ),
        Invalid(
            code="""
            from typing import NamedTuple as NT

            class Foo(NT):
                pass
            """,
            expected_replacement="""
            from typing import NamedTuple as NT

            @dataclass(frozen=True)
            class Foo:
                pass
            """,
        ),
        Invalid(
            code="""
            import typing as typ

            class Foo(typ.NamedTuple):
                pass
            """,
            expected_replacement="""
            import typing as typ

            @dataclass(frozen=True)
            class Foo:
                pass
            """,
        ),
        Invalid(
            code="""
            from typing import NamedTuple

            class Foo(NamedTuple, AnotherBase, YetAnotherBase):
                pass
            """,
            expected_replacement="""
            from typing import NamedTuple

            @dataclass(frozen=True)
            class Foo(AnotherBase, YetAnotherBase):
                pass
            """,
        ),
        Invalid(
            code="""
            from typing import NamedTuple

            class OuterClass(SomeBase):
                class InnerClass(NamedTuple):
                    pass
            """,
            expected_replacement="""
            from typing import NamedTuple

            class OuterClass(SomeBase):
                @dataclass(frozen=True)
                class InnerClass:
                    pass
            """,
        ),
        Invalid(
            code="""
            from typing import NamedTuple

            @some_other_decorator
            class Foo(NamedTuple):
                pass
            """,
            expected_replacement="""
            from typing import NamedTuple

            @some_other_decorator
            @dataclass(frozen=True)
            class Foo:
                pass
            """,
        ),
    ]

    qualified_namedtuple = QualifiedName(
        name="typing.NamedTuple", source=QualifiedNameSource.IMPORT
    )

    def leave_ClassDef(self, original_node: cst.ClassDef) -> None:
        (namedtuple_base, new_bases) = self.partition_bases(original_node.bases)
        if namedtuple_base is not None:
            call = ensure_type(parse_expression("dataclass(frozen=True)"), cst.Call)

            replacement = original_node.with_changes(
                lpar=MaybeSentinel.DEFAULT,
                rpar=MaybeSentinel.DEFAULT,
                bases=new_bases,
                decorators=list(original_node.decorators)
                + [cst.Decorator(decorator=call)],
            )
            self.report(original_node, replacement=replacement)

    def partition_bases(
        self, original_bases: Sequence[cst.Arg]
    ) -> Tuple[Optional[cst.Arg], List[cst.Arg]]:
        # Returns a tuple of NamedTuple base object if it exists, and a list of non-NamedTuple bases
        namedtuple_base: Optional[cst.Arg] = None
        new_bases: List[cst.Arg] = []
        for base_class in original_bases:
            if QualifiedNameProvider.has_name(
                self, base_class.value, self.qualified_namedtuple
            ):
                namedtuple_base = base_class
            else:
                new_bases.append(base_class)
        return (namedtuple_base, new_bases)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# Usage:
#
#   $ python -m fixit.cli.apply_fix --help
#   $ python -m fixit.cli.apply_fix --rules AvoidOrInExceptRule
#   $ python -m fixit.cli.apply_fix . --rules AvoidOrInExceptRule

import argparse
import itertools
import shutil
import sys
import time
from dataclasses import dataclass
from multiprocessing import Manager
from pathlib import Path
from typing import Iterable, List, Mapping, Optional, TYPE_CHECKING

from libcst import parse_module, ParserSyntaxError
from libcst.codemod._cli import invoke_formatter
from libcst.metadata import MetadataWrapper

from fixit.cli import find_files, map_paths
from fixit.cli.args import (
    get_compact_parser,
    get_metadata_cache_parser,
    get_multiprocessing_parser,
    get_paths_parser,
    get_rules_parser,
    get_skip_autoformatter_parser,
    get_skip_ignore_byte_marker_parser,
    get_skip_ignore_comments_parser,
    LintWorkers,
)
from fixit.cli.formatter import LintRuleReportFormatter
from fixit.cli.full_repo_metadata import (
    get_metadata_caches,
    rules_require_metadata_cache,
)
from fixit.cli.utils import print_red
from fixit.common.config import get_lint_config
from fixit.common.report import BaseLintRuleReport
from fixit.common.utils import LintRuleCollectionT
from fixit.rule_lint_engine import (
    lint_file_and_apply_patches,
    LintRuleReportsWithAppliedPatches,
)


if TYPE_CHECKING:
    from libcst.metadata.base_provider import ProviderT

MAX_ITER: int = 100

DESCRIPTION: str = """Runs a lint rule's autofixer over all of over a set of files or
directories. This is similar to the functionality provided by LibCST codemods
(https://libcst.readthedocs.io/en/latest/codemods_tutorial.html), but limited to the
small subset of APIs provided by Fixit."""

PARENTS: List[argparse.ArgumentParser] = [
    get_rules_parser(),
    get_metadata_cache_parser(),
    get_paths_parser(),
    get_skip_ignore_comments_parser(),
    get_skip_ignore_byte_marker_parser(),
    get_skip_autoformatter_parser(),
    get_compact_parser(),
    get_multiprocessing_parser(),
]


@dataclass(frozen=True)
class LintOpts:
    rules: LintRuleCollectionT
    use_ignore_byte_markers: bool
    use_ignore_comments: bool
    skip_autoformatter: bool
    formatter: LintRuleReportFormatter
    patched_files_list: Optional[List[str]] = None


class AutofixingLintRuleReportFormatter(LintRuleReportFormatter):
    def _format_header(self, report: BaseLintRuleReport) -> str:
        fixed_str = " [applied fix]" if report.patch is not None else ""
        return f"{super()._format_header(report)}{fixed_str}"


def get_one_patchable_report_for_path(
    path: Path,
    source: bytes,
    rules: LintRuleCollectionT,
    use_ignore_byte_markers: bool,
    use_ignore_comments: bool,
    metadata_cache: Optional[Mapping["ProviderT", object]],
) -> LintRuleReportsWithAppliedPatches:
    cst_wrapper: Optional[MetadataWrapper] = None
    if metadata_cache is not None:
        cst_wrapper = MetadataWrapper(
            parse_module(source),
            True,
            metadata_cache,
        )

    return lint_file_and_apply_patches(
        path,
        source,
        rules=rules,
        use_ignore_byte_markers=use_ignore_byte_markers,
        use_ignore_comments=use_ignore_comments,
        # We will need to regenerate metadata cache every time a patch is applied.
        max_iter=1,
        cst_wrapper=cst_wrapper,
        find_unused_suppressions=True,
    )


def apply_fix_operation(
    path: Path,
    opts: LintOpts,
    metadata_cache: Optional[Mapping["ProviderT", object]] = None,
) -> Iterable[str]:
    with open(path, "rb") as f:
        source = f.read()
    patched_files_list = opts.patched_files_list
    try:
        if patched_files_list is None:
            lint_result = lint_file_and_apply_patches(
                path,
                source,
                rules=opts.rules,
                use_ignore_byte_markers=opts.use_ignore_byte_markers,
                use_ignore_comments=opts.use_ignore_comments,
                find_unused_suppressions=True,
            )
            raw_reports = lint_result.reports
            updated_source = lint_result.patched_source
            if updated_source != source:
                if not opts.skip_autoformatter:
                    # Format the code using the config file's formatter.
                    updated_source = invoke_formatter(
                        get_lint_config().formatter, updated_source
                    )
                with open(path, "w") as f:
                    f.write(updated_source)
        else:
            lint_result = get_one_patchable_report_for_path(
                path,
                source,
                opts.rules,
                opts.use_ignore_byte_markers,
                opts.use_ignore_comments,
                metadata_cache,
            )
            # Will either be a single patchable report, or a collection of non-patchable reports.
            raw_reports = lint_result.reports
            updated_source = lint_result.patched_source
            if updated_source != source:
                # We don't do any formatting here as it's wasteful. The caller should handle formatting all files at the end.
                with open(path, "w") as f:
                    f.write(updated_source)

                patched_files_list.append(str(path))
                # Return only the report that was used in the patched source.
                return [next(opts.formatter.format(rr) for rr in raw_reports)]
    except (SyntaxError, ParserSyntaxError) as e:
        print_red(
            f"Encountered the following error while parsing source code in file {path}:"
        )
        print(e)
        return []

    return [opts.formatter.format(rr) for rr in raw_reports]


def call_map_paths_and_print_reports(
    file_paths: Iterable[str],
    opts: LintOpts,
    workers: LintWorkers,
    metadata_caches: Optional[Mapping[str, Mapping["ProviderT", object]]] = None,
) -> int:
    """
    Calls map_paths with `apply_fix_operation`, and the passed in file_paths, opts, workers, metadata_cache.
    Returns the number of reports in the files.
    """
    num_reports = 0
    formatted_reports_iter = itertools.chain.from_iterable(
        map_paths(
            apply_fix_operation,
            file_paths,
            opts,
            workers=workers,
            metadata_caches=metadata_caches,
        )
    )

    for formatted_report in formatted_reports_iter:
        # Reports are yielded as soon as they're available. Stream the output to the
        # terminal.
        print(formatted_report)
        num_reports += 1
    return num_reports


def _add_arguments(parser: argparse.ArgumentParser) -> None:
    """All required arguments for `apply_fix`"""
    # Add arguments for `apply fix`
    pass


def register_subparser(parser: Optional[argparse._SubParsersAction] = None) -> None:
    """Add parser or subparser for `apply_fix` command."""
    if parser is None:
        apply_fix_parser = argparse.ArgumentParser(
            description=DESCRIPTION,
            parents=PARENTS,
        )
        _add_arguments(apply_fix_parser)
        sys.exit(_main(apply_fix_parser.parse_args()))

    else:
        apply_fix_parser = parser.add_parser(
            "apply_fix",
            description=DESCRIPTION,
            parents=PARENTS,
            help="Apply lint rule fix(s)",
        )
        _add_arguments(apply_fix_parser)
        apply_fix_parser.set_defaults(subparser_fn=_main)


def _main(args: argparse.Namespace) -> int:
    width = shutil.get_terminal_size(fallback=(80, 24)).columns

    rules = args.rules
    use_ignore_byte_markers = args.use_ignore_byte_markers
    use_ignore_comments = args.use_ignore_comments
    skip_autoformatter = args.skip_autoformatter
    formatter = AutofixingLintRuleReportFormatter(width, args.compact)
    workers = args.workers

    # Find files if directory was provided.
    file_paths = tuple(find_files(args.paths))

    if not args.compact:
        print(f"Scanning {len(file_paths)} files")
        print("\n".join(file_paths))
        print()
    start_time = time.time()

    total_reports_count = 0

    if rules_require_metadata_cache(rules):
        touched_files = set()
        next_files = file_paths
        with Manager() as manager:
            # Avoid getting stuck in an infinite loop.
            for _ in range(MAX_ITER):
                if not next_files:
                    break

                patched_files = manager.list()
                metadata_caches = get_metadata_caches(args.cache_timeout, next_files)

                next_files = []
                # opts is a more type-safe version of args that we pass around
                opts = LintOpts(
                    rules=rules,
                    use_ignore_byte_markers=use_ignore_byte_markers,
                    use_ignore_comments=use_ignore_comments,
                    skip_autoformatter=skip_autoformatter,
                    formatter=formatter,
                    patched_files_list=patched_files,
                )
                total_reports_count += call_map_paths_and_print_reports(
                    next_files, opts, workers, metadata_caches
                )
                next_files = list(patched_files)
                touched_files.update(patched_files)

        # Finally, format all the touched files.
        if not skip_autoformatter:
            for path in touched_files:
                with open(path, "rb") as f:
                    source = f.read()
                # Format the code using the config file's formatter.
                formatted_source = invoke_formatter(get_lint_config().formatter, source)
                with open(path, "wb") as f:
                    f.write(formatted_source)

    else:
        # opts is a more type-safe version of args that we pass around
        opts = LintOpts(
            rules=rules,
            use_ignore_byte_markers=use_ignore_byte_markers,
            use_ignore_comments=use_ignore_comments,
            skip_autoformatter=skip_autoformatter,
            formatter=formatter,
        )

        total_reports_count = call_map_paths_and_print_reports(
            file_paths, opts, workers, None
        )

    if not args.compact:
        print()
        print(
            f"Found {total_reports_count} reports in {len(file_paths)} files in "
            + f"{time.time() - start_time :.2f} seconds."
        )

    return 0


if __name__ == "__main__":
    register_subparser()



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from pathlib import Path

from fixit import add_lint_rule_tests_to_module
from fixit.common.base import LintConfig
from fixit.common.config import get_lint_config, get_rules_from_config


# Add all the CstLintRules from `fixit.rules` package to this module as unit tests.
CONFIG: LintConfig = get_lint_config()
add_lint_rule_tests_to_module(
    globals(),
    get_rules_from_config(),
    fixture_dir=Path(CONFIG.fixture_dir),
    rules_package="fixit.rules",
)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import cast, Collection, List, Sequence

import libcst as cst
from libcst.metadata import ParentNodeProvider, PositionProvider

from fixit.common.base import CstContext, CstLintRule
from fixit.common.ignores import AllRulesType, IgnoreInfo, SuppressionComment
from fixit.common.insert_suppressions import (
    DEFAULT_MIN_COMMENT_WIDTH,
    SuppressionComment as NewSuppressionComment,
    SuppressionCommentKind,
)


UNUSED_SUPPRESSION_COMMENT_MESSAGE = "Unused lint suppression. This comment is not suppressing lint errors and should be removed."
UNUSED_SUPPRESSION_CODES_IN_COMMENT_MESSAGE = "The codes `{lint_codes}` in this comment are not suppressing lint errors. They can be removed from this suppression."


def _compose_new_comment(
    local_supp_comment: SuppressionComment,
    unneeded_codes: Collection[str],
    comment_physical_line: int,
) -> Sequence[str]:
    ignored_rules = local_supp_comment.ignored_rules
    # Just a check for the type-checker - it should never actually be AllRulesType
    # because we don't support that in lint-fixme/lint-ignore comments.
    # TODO: We can remove the AllRulesType check once we deprecate noqa.
    if isinstance(ignored_rules, AllRulesType):
        raise ValueError(
            f"Got `AllRulesType` for the suppressed rules in comment on line {comment_physical_line}"
        )

    new_codes = ", ".join([c for c in ignored_rules if c not in unneeded_codes])
    if not new_codes:
        return []

    kind = (
        SuppressionCommentKind.FIXME
        if local_supp_comment.kind == "fixme"
        else SuppressionCommentKind.IGNORE
    )
    first_line = local_supp_comment.tokens[0]
    width = max(DEFAULT_MIN_COMMENT_WIDTH, len(first_line.string))

    # Construct a new comment.
    new_lines = NewSuppressionComment(
        kind=kind,
        before_line=comment_physical_line + 1,
        code=new_codes,
        message=local_supp_comment.reason,
        max_lines=len(local_supp_comment.tokens),
    ).to_lines(width)

    return new_lines


def _get_unused_codes_in_comment(
    local_supp_comment: SuppressionComment, ignored_rules_that_ran: Collection[str]
) -> Collection[str]:
    """Returns a subset of the rules in the comment which did not show up in any report."""
    return {
        ir
        for ir in ignored_rules_that_ran
        if not any(ir == report.code for report in local_supp_comment.used_by)
    }


def _modify_parent_attribute(
    parent_node: cst.CSTNode,
    attribute_name: str,
    idx_left: int,
    idx_right: int,
    list_to_insert: List[cst.EmptyLine],
) -> List[cst.EmptyLine]:
    """
    Replace the section starting at idx_left and ending at idx_right of the Sequence[cst.EmptyLine]-type attribute
    with `list_to_insert`.
    Returns the replacement attribute.
    """
    attribute_value = getattr(parent_node, attribute_name)
    attribute_value = cast(Sequence[cst.EmptyLine], attribute_value)
    return (
        list(attribute_value[:idx_left])
        + list_to_insert
        + list(attribute_value[idx_right:])
    )


# This is a special lint rule that should not be included in the fixit.rules package as it needs to run after all other rules.
class RemoveUnusedSuppressionsRule(CstLintRule):
    """
    The rule is run by `lint_file` after all other rules have finished running.
    It requires a completed reports dict in order to work properly.
    """

    METADATA_DEPENDENCIES = (ParentNodeProvider,)

    def __init__(
        self,
        context: CstContext,
    ) -> None:
        # TODO: Support Flake8 suppressions.
        super().__init__(context)
        ignore_info: object = self.context.config.rule_config[self.__class__.__name__][
            "ignore_info"
        ]
        rules: object = self.context.config.rule_config[self.__class__.__name__][
            "rules"
        ]
        assert isinstance(ignore_info, IgnoreInfo)
        assert isinstance(rules, list)

        self.ignore_info: IgnoreInfo = ignore_info
        self.rule_names: Collection[str] = {r.__name__ for r in rules}

    def _handle_emptyline_sequence_attribute(
        self, node: cst.CSTNode, attribute_name: str
    ) -> None:
        for idx, emptyline_node in enumerate(getattr(node, attribute_name)):
            comment = emptyline_node.comment
            if comment is None:
                continue
            comment_physical_line = self.get_metadata(
                PositionProvider, comment
            ).start.line
            local_supp_comments: Sequence[
                SuppressionComment
            ] = self.ignore_info.local_ignore_info.local_suppression_comments_by_line.get(
                comment_physical_line, []
            )
            if local_supp_comments:
                # We have found a `lint-ignore` or `lint-fixme` comment.
                # It should only ever contain one element since a comment's 'local comment' will only be itself.
                assert len(local_supp_comments) == 1
                local_supp_comment = local_supp_comments[0]
                if local_supp_comment.tokens[0].string == comment.value:
                    # We only call _handle_suppression_comment on the first line of the comment, it'll handle the remaining lines.
                    self._handle_suppression_comment(
                        node,
                        attribute_name,
                        idx,
                        local_supp_comment,
                        comment_physical_line,
                    )

    def _handle_node_with_leading_lines(self, node: cst.CSTNode) -> None:
        # pyre-fixme[16]: `CSTNode` has no attribute `leading_lines`.
        if hasattr(node, "leading_lines") and getattr(node, "leading_lines"):
            self._handle_emptyline_sequence_attribute(node, "leading_lines")

    def visit_Module(self, node: cst.Module) -> None:
        if node.header:
            self._handle_emptyline_sequence_attribute(node, "header")
        if node.footer:
            self._handle_emptyline_sequence_attribute(node, "footer")

    def visit_SimpleStatementLine(self, node: cst.SimpleStatementLine) -> None:
        self._handle_node_with_leading_lines(node)

    def visit_If(self, node: cst.If) -> None:
        self._handle_node_with_leading_lines(node)

    def visit_Else(self, node: cst.Else) -> None:
        self._handle_node_with_leading_lines(node)

    def visit_BaseCompoundStatement(self, node: cst.BaseCompoundStatement) -> None:
        self._handle_node_with_leading_lines(node)

    def visit_ExceptHandler(self, node: cst.ExceptHandler) -> None:
        self._handle_node_with_leading_lines(node)

    def visit_Finally(self, node: cst.Finally) -> None:
        self._handle_node_with_leading_lines(node)

    def visit_Try(self, node: cst.Try) -> None:
        self._handle_node_with_leading_lines(node)

    def visit_Decorator(self, node: cst.Decorator) -> None:
        self._handle_node_with_leading_lines(node)

    def visit_FunctionDef(self, node: cst.FunctionDef) -> None:
        if node.leading_lines:
            self._handle_emptyline_sequence_attribute(node, "leading_lines")
        if node.lines_after_decorators:
            self._handle_emptyline_sequence_attribute(node, "lines_after_decorators")

    def visit_ClassDef(self, node: cst.ClassDef) -> None:
        if node.leading_lines:
            self._handle_emptyline_sequence_attribute(node, "leading_lines")
        if node.lines_after_decorators:
            self._handle_emptyline_sequence_attribute(node, "lines_after_decorators")

    def visit_With(self, node: cst.With) -> None:
        self._handle_node_with_leading_lines(node)

    def visit_For(self, node: cst.For) -> None:
        self._handle_node_with_leading_lines(node)

    def visit_While(self, node: cst.While) -> None:
        self._handle_node_with_leading_lines(node)

    def visit_ParenthesizedWhitespace(self, node: cst.ParenthesizedWhitespace) -> None:
        if node.empty_lines:
            self._handle_emptyline_sequence_attribute(node, "empty_lines")

    def _handle_suppression_comment(
        self,
        parent_node: cst.CSTNode,
        parent_attribute_name: str,
        index: int,
        local_supp_comment: SuppressionComment,
        comment_physical_line: int,
    ) -> None:
        ignored_rules = local_supp_comment.ignored_rules
        # Just a check for the type-checker - it should never actually be AllRulesType
        # because we don't support that in lint-fixme/lint-ignore comments.
        # TODO: We can remove the AllRulesType check once we deprecate noqa.
        if isinstance(ignored_rules, AllRulesType):
            return

        # First find the suppressed rules that were included in the lint run. If a rule was not included
        # in this run, we CANNOT know for sure that this lint suppression is unused.
        ignored_rules_that_ran = {ig for ig in ignored_rules if ig in self.rule_names}

        if not ignored_rules_that_ran:
            return

        lines_span = len(local_supp_comment.tokens)

        unneeded_codes = _get_unused_codes_in_comment(
            local_supp_comment, ignored_rules_that_ran
        )
        if unneeded_codes:
            new_comment_lines = _compose_new_comment(
                local_supp_comment, unneeded_codes, comment_physical_line
            )
            if not new_comment_lines:
                # If we're here, all of the codes in this suppression refer to rules that ran, and
                # none of comment's codes suppress anything, so we report this comment and offer to remove it.
                new_parent_attribute_value = _modify_parent_attribute(
                    parent_node, parent_attribute_name, index, index + lines_span, []
                )
                self.report(
                    parent_node,
                    message=UNUSED_SUPPRESSION_COMMENT_MESSAGE,
                    replacement=parent_node.with_changes(
                        **{parent_attribute_name: new_parent_attribute_value}
                    ),
                )
            else:
                node_to_replace = getattr(parent_node, parent_attribute_name)[index]
                replacement_emptyline_nodes = [
                    cst.EmptyLine(
                        indent=node_to_replace.indent,
                        whitespace=node_to_replace.whitespace,
                        comment=cst.Comment(line),
                    )
                    for line in new_comment_lines
                ]
                new_parent_attribute_value: List[
                    cst.EmptyLine
                ] = _modify_parent_attribute(
                    parent_node,
                    parent_attribute_name,
                    index,
                    index + lines_span,
                    replacement_emptyline_nodes,
                )

                self.report(
                    parent_node,
                    message=UNUSED_SUPPRESSION_CODES_IN_COMMENT_MESSAGE.format(
                        lint_codes="`, `".join(uc for uc in unneeded_codes)
                    ),
                    replacement=parent_node.with_changes(
                        **{parent_attribute_name: new_parent_attribute_value}
                    ),
                )



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import List

import libcst as cst

from fixit import CstLintRule, InvalidTestCase as Invalid, ValidTestCase as Valid


class UsePlusForStringConcatRule(CstLintRule):
    """
    Enforces use of explicit string concatenations using a ``+`` sign where an implicit concatenation is detected.
    An implicit concatenation is a tuple or a call with multiple strings and a missing comma, e.g: ``("a" "b")``, and may have unexpected results.
    """

    MESSAGE: str = (
        "Implicit string concatenation detected, please add '+' to be explicit. "
        + 'E.g. a tuple or a call ("a" "b") with a missing comma results in multiple strings '
        + "being concatenated as one string and causes unexpected behaviour."
    )
    VALID = [Valid("'abc'"), Valid("'abc' + 'def'"), Valid("f'abc'")]
    INVALID = [
        Invalid(
            "'ab' 'cd'",
            expected_replacement="('ab' + 'cd')",
        ),
        # We can deal with nested concatenated strings
        Invalid(
            "'ab' 'cd' 'ef' 'gh'",
            expected_replacement="('ab' + 'cd' + 'ef' + 'gh')",
        ),
        # works for f-strings too
        Invalid(
            "f'ab' f'cd'",
            expected_replacement="(f'ab' + f'cd')",
        ),
        # arbitrary whitespace between the elements is preserved
        Invalid(
            """
                (
                    # comment
                    'ab'  # middle comment
                    'cd'  # trailing comment
                )
            """,
            expected_replacement="""
                (
                    # comment
                    'ab'  # middle comment
                    + 'cd'  # trailing comment
                )
            """,
        ),
    ]

    def visit_ConcatenatedString(self, node: cst.ConcatenatedString) -> None:
        # Skip if our immediate parent is also a ConcatenatedString, since our parent
        # should've already reported this violation.
        if isinstance(self.context.node_stack[-2], cst.ConcatenatedString):
            return

        # collect nested ConcatenatedString nodes into a flat list from outer to
        # innermost children
        children: List[cst.ConcatenatedString] = []
        el = node
        while isinstance(el, cst.ConcatenatedString):
            children.append(el)
            # left cannot be a ConcatenatedString, only right can.
            el = el.right

        # Build up a replacement by starting with the innermost child
        replacement = children[-1].right
        for el in reversed(children):
            replacement = cst.BinaryOperation(
                left=el.left,  # left is never a ConcatenatedString
                operator=cst.Add(
                    whitespace_before=el.whitespace_between,
                    whitespace_after=cst.SimpleWhitespace(" "),
                ),
                right=replacement,
                lpar=el.lpar,
                rpar=el.rpar,
            )

        # A binary operation has a lower priority in the order-of-operations than an
        # implicitly concatenated string, so we need to make sure the replacement is
        # parenthesized to make our change safe.
        if not replacement.lpar:
            # There's a good chance that the formatting might be messed up by this, but
            # black should be able to sort it out when it gets run next time.
            #
            # Because of the changes needed (e.g. increased indentation of children),
            # it's not really sane/possible for us to format this any better.
            replacement = replacement.with_changes(
                lpar=[cst.LeftParen()], rpar=[cst.RightParen()]
            )

        self.report(node, replacement=replacement)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import libcst as cst
import libcst.matchers as m
from libcst.helpers import get_full_name_for_node

from fixit import CstLintRule, InvalidTestCase as Invalid, ValidTestCase as Valid


UNNECESSARY_LAMBDA: str = (
    "The lambda that is wrapping {function} is redundant. "
    + "It can unwrapped safely and used purely."
)


class NoRedundantLambdaRule(CstLintRule):
    """A lamba function which has a single objective of
    passing all it is arguments to another callable can
    be safely replaced by that callable."""

    VALID = [
        Valid("lambda x: foo(y)"),
        Valid("lambda x: foo(x, y)"),
        Valid("lambda x, y: foo(x)"),
        Valid("lambda *, x: foo(x)"),
        Valid("lambda x = y: foo(x)"),
        Valid("lambda x, y: foo(y, x)"),
        Valid("lambda self: self.func()"),
        Valid("lambda x, y: foo(y=x, x=y)"),
        Valid("lambda x, y, *z: foo(x, y, z)"),
        Valid("lambda x, y, **z: foo(x, y, z)"),
    ]
    INVALID = [
        Invalid("lambda: self.func()", expected_replacement="self.func"),
        Invalid("lambda x: foo(x)", expected_replacement="foo"),
        Invalid(
            "lambda x, y, z: (t + u).math_call(x, y, z)",
            expected_replacement="(t + u).math_call",
        ),
    ]

    @staticmethod
    def _is_simple_parameter_spec(node: cst.Parameters) -> bool:
        if (
            node.star_kwarg is not None
            or len(node.kwonly_params) > 0
            or len(node.posonly_params) > 0
            or not isinstance(node.star_arg, cst.MaybeSentinel)
        ):
            return False

        return all(param.default is None for param in node.params)

    def visit_Lambda(self, node: cst.Lambda) -> None:
        if m.matches(
            node,
            m.Lambda(
                params=m.MatchIfTrue(self._is_simple_parameter_spec),
                body=m.Call(
                    args=[
                        m.Arg(
                            value=m.Name(value=param.name.value), star="", keyword=None
                        )
                        for param in node.params.params
                    ]
                ),
            ),
        ):
            call = cst.ensure_type(node.body, cst.Call)
            full_name = get_full_name_for_node(call)
            if full_name is None:
                full_name = "function"

            self.report(
                node,
                UNNECESSARY_LAMBDA.format(function=full_name),
                replacement=call.func,
            )



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import importlib
import inspect
import json
import pkgutil
import re
import textwrap
from dataclasses import dataclass
from pathlib import Path
from types import ModuleType
from typing import cast, Dict, List, Optional, Set, Type, Union

import libcst as cst
from libcst._add_slots import add_slots
from libcst.metadata import BaseMetadataProvider, MetadataWrapper, TypeInferenceProvider
from libcst.metadata.type_inference_provider import PyreData

from fixit.common.base import CstLintRule, LintConfig, LintRuleT
from fixit.common.pseudo_rule import PseudoLintRule


def _dedent(src: str) -> str:
    src = re.sub(r"\A\n", "", src)
    return textwrap.dedent(src)


def dedent_with_lstrip(src: str) -> str:
    src = textwrap.dedent(src)
    if src.startswith("\n"):
        return "".join(src[1:])
    return src


def _str_or_any(value: Optional[int]) -> str:
    return "<any>" if value is None else str(value)


class DuplicateLintRuleNameError(Exception):
    pass


class FixtureFileNotFoundError(Exception):
    pass


class LintRuleNotFoundError(Exception):
    pass


LintRuleCollectionT = Set[Union[Type[CstLintRule], Type[PseudoLintRule]]]
DEFAULT_FILENAME: str = "not/a/real/file/path.py"
DEFAULT_CONFIG: LintConfig = LintConfig(
    repo_root=str(
        Path(__file__).parent.parent
    ),  # Set base config repo_root to `fixit` directory for testing.
)


@add_slots
@dataclass(frozen=True)
class ValidTestCase:
    code: str
    filename: str = DEFAULT_FILENAME
    config: LintConfig = DEFAULT_CONFIG


@add_slots
@dataclass(frozen=True)
class InvalidTestCase:
    code: str
    kind: Optional[str] = None
    line: Optional[int] = None
    column: Optional[int] = None
    expected_replacement: Optional[str] = None
    filename: str = DEFAULT_FILENAME
    config: LintConfig = DEFAULT_CONFIG
    expected_message: Optional[str] = None

    @property
    def expected_str(self) -> str:
        return f"{_str_or_any(self.line)}:{_str_or_any(self.column)}: {self.kind} ..."


def import_submodules(package: str, recursive: bool = True) -> Dict[str, ModuleType]:
    """Import all submodules of a module, recursively, including subpackages."""
    # pyre-fixme[35]: Target cannot be annotated.
    package: ModuleType = importlib.import_module(package)
    results = {}
    # pyre-fixme[16]: `ModuleType` has no attribute `__path__`.
    for _loader, name, is_pkg in pkgutil.walk_packages(package.__path__):
        full_name = package.__name__ + "." + name
        try:
            results[full_name] = importlib.import_module(full_name)
        except ModuleNotFoundError:
            pass
        if recursive and is_pkg:
            results.update(import_submodules(full_name))
    return results


def import_distinct_rules_from_package(
    package: str,
    block_list_rules: List[str] = [],
    seen_names: Optional[Set[str]] = None,
    allow_list_rules: Optional[List[str]] = None,
) -> LintRuleCollectionT:
    # Import all rules from the specified package, omitting rules that appear in the block list.
    # Raises error on repeated rule names.
    # Optional parameter `seen_names` accepts set of names that should not occur in this package.
    rules: LintRuleCollectionT = set()
    if seen_names is None:
        # pyre-fixme[35]: Target cannot be annotated.
        seen_names: Set[str] = set()
    for _module_name, module in import_submodules(package).items():
        for name in dir(module):
            try:
                obj = getattr(module, name)
                if (
                    obj is not CstLintRule
                    and (
                        issubclass(obj, CstLintRule) or issubclass(obj, PseudoLintRule)
                    )
                    and not inspect.isabstract(obj)
                ):
                    if name in seen_names:
                        raise DuplicateLintRuleNameError(
                            f"Lint rule name {name} is duplicated."
                        )
                    # Add all names (even block-listed ones) to the `names` set for duplicate checking.
                    seen_names.add(name)
                    # For backwards compatibility if `allow_list_rules` is missing fall back to all allowed
                    if not allow_list_rules or name in allow_list_rules:
                        if name not in block_list_rules:
                            rules.add(obj)
            except TypeError:
                continue
    return rules


def gen_type_inference_wrapper(code: str, pyre_fixture_path: Path) -> MetadataWrapper:
    # Given test case source code and a path to a pyre fixture file, generate a MetadataWrapper for a lint rule test case.
    module = cst.parse_module(_dedent(code))
    provider_type = TypeInferenceProvider
    try:
        pyre_json_data: PyreData = json.loads(pyre_fixture_path.read_text())
    except FileNotFoundError as e:
        raise FixtureFileNotFoundError(
            f"Fixture file not found at {e.filename}. "
            + "Please run `python -m fixit.common.generate_pyre_fixtures <rule>` to generate fixtures."
        )
    return MetadataWrapper(
        module=module,
        cache={cast(Type[BaseMetadataProvider[object]], provider_type): pyre_json_data},
    )


def import_rule_from_package(
    package_name: str, rule_class_name: str
) -> Optional[LintRuleT]:
    # Imports the first rule with matching class name found in specified package.
    rule: Optional[LintRuleT] = None
    package = importlib.import_module(package_name)
    for _loader, name, is_pkg in pkgutil.walk_packages(
        getattr(package, "__path__", None)
    ):
        full_package_or_module_name = package.__name__ + "." + name
        try:
            module = importlib.import_module(full_package_or_module_name)
            rule = getattr(module, rule_class_name, None)
        except ModuleNotFoundError:
            pass
        if is_pkg:
            rule = import_rule_from_package(
                full_package_or_module_name, rule_class_name
            )

        if rule is not None:
            # Stop early if we have found the rule.
            return rule
    return rule


def find_and_import_rule(rule_class_name: str, packages: List[str]) -> LintRuleT:
    for package in packages:
        imported_rule = import_rule_from_package(package, rule_class_name)
        if imported_rule is not None:
            return imported_rule

    # If we get here, the rule was not found.
    raise LintRuleNotFoundError(
        f"Could not find lint rule {rule_class_name} in the following packages: \n"
        + "\n".join(packages)
    )



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import libcst as cst
import libcst.matchers as m

from fixit import CstLintRule, InvalidTestCase as Invalid, ValidTestCase as Valid


UNNECESSARY_GENERATOR: str = (
    "It's unnecessary to use {func} around a generator expression, since "
    + "there are equivalent comprehensions for this type."
)
UNNECESSARY_LIST_COMPREHENSION: str = (
    "It's unnecessary to use a list comprehension inside a call to {func} "
    + "since there are equivalent comprehensions for this type"
)


class RewriteToComprehensionRule(CstLintRule):
    """
    A derivative of flake8-comprehensions's C400-C402 and C403-C404.
    Comprehensions are more efficient than functions calls. This C400-C402
    suggest to use `dict/set/list` comprehensions rather than respective
    function calls whenever possible. C403-C404 suggest to remove unnecessary
    list comprehension in a set/dict call, and replace it with set/dict
    comprehension.
    """

    VALID = [
        Valid("[val for val in iterable]"),
        Valid("{val for val in iterable}"),
        Valid("{val: val+1 for val in iterable}"),
        # A function call is valid if the elt is a function that returns a tuple.
        Valid("dict(line.strip().split('=', 1) for line in attr_file)"),
    ]

    INVALID = [
        Invalid(
            "list(val for val in iterable)",
            expected_replacement="[val for val in iterable]",
        ),
        # Nested list comprehenstion
        Invalid(
            "list(val for row in matrix for val in row)",
            expected_replacement="[val for row in matrix for val in row]",
        ),
        Invalid(
            "set(val for val in iterable)",
            expected_replacement="{val for val in iterable}",
        ),
        Invalid(
            "dict((x, f(x)) for val in iterable)",
            expected_replacement="{x: f(x) for val in iterable}",
        ),
        Invalid(
            "dict((x, y) for y, x in iterable)",
            expected_replacement="{x: y for y, x in iterable}",
        ),
        Invalid(
            "dict([val, val+1] for val in iterable)",
            expected_replacement="{val: val+1 for val in iterable}",
        ),
        Invalid(
            'dict((x["name"], json.loads(x["data"])) for x in responses)',
            expected_replacement='{x["name"]: json.loads(x["data"]) for x in responses}',
        ),
        # Nested dict comprehension
        Invalid(
            "dict((k, v) for k, v in iter for iter in iters)",
            expected_replacement="{k: v for k, v in iter for iter in iters}",
        ),
        Invalid(
            "set([val for val in iterable])",
            expected_replacement="{val for val in iterable}",
        ),
        Invalid(
            "dict([[val, val+1] for val in iterable])",
            expected_replacement="{val: val+1 for val in iterable}",
        ),
        Invalid(
            "dict([(x, f(x)) for x in foo])",
            expected_replacement="{x: f(x) for x in foo}",
        ),
        Invalid(
            "dict([(x, y) for y, x in iterable])",
            expected_replacement="{x: y for y, x in iterable}",
        ),
        Invalid(
            "set([val for row in matrix for val in row])",
            expected_replacement="{val for row in matrix for val in row}",
        ),
    ]

    def visit_Call(self, node: cst.Call) -> None:
        if m.matches(
            node,
            m.Call(
                func=m.Name("list") | m.Name("set") | m.Name("dict"),
                args=[m.Arg(value=m.GeneratorExp() | m.ListComp())],
            ),
        ):
            call_name = cst.ensure_type(node.func, cst.Name).value

            if m.matches(node.args[0].value, m.GeneratorExp()):
                exp = cst.ensure_type(node.args[0].value, cst.GeneratorExp)
                message_formatter = UNNECESSARY_GENERATOR
            else:
                exp = cst.ensure_type(node.args[0].value, cst.ListComp)
                message_formatter = UNNECESSARY_LIST_COMPREHENSION

            replacement = None
            if call_name == "list":
                replacement = node.deep_replace(
                    node, cst.ListComp(elt=exp.elt, for_in=exp.for_in)
                )
            elif call_name == "set":
                replacement = node.deep_replace(
                    node, cst.SetComp(elt=exp.elt, for_in=exp.for_in)
                )
            elif call_name == "dict":
                elt = exp.elt
                key = None
                value = None
                if m.matches(elt, m.Tuple(m.DoNotCare(), m.DoNotCare())):
                    elt = cst.ensure_type(elt, cst.Tuple)
                    key = elt.elements[0].value
                    value = elt.elements[1].value
                elif m.matches(elt, m.List(m.DoNotCare(), m.DoNotCare())):
                    elt = cst.ensure_type(elt, cst.List)
                    key = elt.elements[0].value
                    value = elt.elements[1].value
                else:
                    # Unrecoginized form
                    return

                replacement = node.deep_replace(
                    node,
                    # pyre-fixme[6]: Expected `BaseAssignTargetExpression` for 1st
                    #  param but got `BaseExpression`.
                    cst.DictComp(key=key, value=value, for_in=exp.for_in),
                )

            self.report(
                node, message_formatter.format(func=call_name), replacement=replacement
            )



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import libcst as cst
import libcst.matchers as m

from fixit import CstLintRule, InvalidTestCase as Invalid, ValidTestCase as Valid


class NoRedundantFStringRule(CstLintRule):
    """
    Remove redundant f-string without placeholders.
    """

    MESSAGE: str = "f-string doesn't have placeholders, remove redundant f-string."

    VALID = [
        Valid('good: str = "good"'),
        Valid('good: str = f"with_arg{arg}"'),
        Valid('good = "good{arg1}".format(1234)'),
        Valid('good = "good".format()'),
        Valid('good = "good" % {}'),
        Valid('good = "good" % ()'),
        Valid('good = rf"good\t+{bar}"'),
    ]

    INVALID = [
        Invalid(
            'bad: str = f"bad" + "bad"',
            line=1,
            expected_replacement='bad: str = "bad" + "bad"',
        ),
        Invalid(
            "bad: str = f'bad'",
            line=1,
            expected_replacement="bad: str = 'bad'",
        ),
        Invalid(
            "bad: str = rf'bad\t+'",
            line=1,
            expected_replacement="bad: str = r'bad\t+'",
        ),
        Invalid(
            'bad: str = f"no args but messing up {{ braces }}"',
            line=1,
            expected_replacement='bad: str = "no args but messing up { braces }"',
        ),
    ]

    def visit_FormattedString(self, node: cst.FormattedString) -> None:
        if not m.matches(node, m.FormattedString(parts=(m.FormattedStringText(),))):
            return

        old_string_inner = cst.ensure_type(node.parts[0], cst.FormattedStringText).value
        if "{{" in old_string_inner or "}}" in old_string_inner:
            old_string_inner = old_string_inner.replace("{{", "{").replace("}}", "}")

        new_string_literal = (
            node.start.replace("f", "").replace("F", "") + old_string_inner + node.end
        )

        self.report(node, replacement=cst.SimpleString(new_string_literal))
